
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 *
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: ipfw_wrap.c
 * Function: Source file of ipfw wrapper of the wireconf library
 *
 * Authors: Junya Nakata, Razvan Beuran
 *
 ***********************************************************************/

#include <sys/queue.h>
#include <sys/socket.h>

#include <net/if.h>
//#include <netinet/in.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <inttypes.h>

#include <time.h>

#include "wireconf.h"


/////////////////////////////////////////////
// Special defines
/////////////////////////////////////////////

// this define is only used in conjunction with the new modified
// version of dummynet that is still in experimental phase 

//#define NEW_DUMMYNET #(version 6??)


///////////////////////////////////////////////
// Special functions
///////////////////////////////////////////////

// handle uint64 alignment (from ip_fw2.c)
static __inline uint64_t
align_uint64 (uint64_t * pll)
{
  uint64_t ret;

  bcopy (pll, &ret, sizeof (ret));
  return ret;
}


///////////////////////////////////////////////
// IPv4 address manipulation
///////////////////////////////////////////////

// convert a string containing an IPv4 address followed by an optional
// port (separated by ':') to an IPv4 data structure;
// return SUCCESS on success, ERROR on error
static int
atoaddr (char *array_address, union in4_addr * ipv4_address, uint16_t * port)
{
  int c, o;

  c = 0;
  o = 0;
  ipv4_address->word = 0;
  *port = 0;

  for (;;)
    {
      if (array_address[c] == '.')
	{
	  c++, o++;
	  continue;
	}
      else if (isdigit (array_address[c]) == 0)
	break;

      ipv4_address->octet[o] = ipv4_address->octet[o] * 10
	+ array_address[c] - '0';
      c++;
    }

  if (array_address[c] == '\0')
    return SUCCESS;
  else if ((o != 3) || (array_address[c] != ':'))
    return ERROR;
  c++;

  for (;;)
    {
      if (isdigit (array_address[c]) == 0)
	break;
      *port = *port * 10 + array_address[c] - '0';
      c++;
    }
  return SUCCESS;
}


/////////////////////////////////////////////
// Socket manipulation functions
/////////////////////////////////////////////

// get a socket identifier;
// return socket identifier on success, ERROR on error
int
get_socket (void)
{
  int socket_id;

  if ((socket_id = socket (AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
    {
      WARNING ("Error creating socket");
      perror ("socket");
      return ERROR;
    }

  return socket_id;
}

// close socket specified by socket_id
void
close_socket (int socket_id)
{
  close (socket_id);
}


////////////////////////////////////////////////
// ipfw rule related functions
////////////////////////////////////////////////

// add an ipfw rule containing a dummynet pipe to the firewall;
// rule number is 'rulenum', pipe number is 'pipe_nr', source and 
// destination are 'src' and 'dst', direction is 'direction';
// return SUCCESS on success, ERROR on error

// NOTE: both 'src' and 'dst' can be either specific 
// IP addresses or the string "any"

int
add_rule_and_pipe (int socket_id, uint16_t rulenum, int pipe_nr,
		   char *src, char *dst, int direction)
{
  // NOTE: debugging info on socket errors can be obtained using 'dmesg'

  // position in command array
  int cmd_pos;	

  // IP address structure
  union in4_addr addr;

  // source and destination ports
  uint16_t src_port, dst_port;
  ipfw_insn cmd[255];
  struct ip_fw *rule;
  socklen_t rule_len;

  // initialize position in command array
  cmd_pos = 0;

  // NOTE: source and destination addresses are processed before
  // source and destination ports for compatibility with old versions of
  // Dummynet, although the most recent one seems to support the
  // interleaved processing as well

  // process source address
  if (strncmp (src, "any", strlen (src)) == 0)
    {
      // source address was "any"
      cmd[cmd_pos].opcode = O_IP_SRC;
      cmd[cmd_pos].len = 0;
      cmd[cmd_pos].arg1 = 0;
      DEBUG ("Cmd src any: opcode=%d len=%d arg1=%d",
	     cmd[cmd_pos].opcode, cmd[cmd_pos].len, cmd[cmd_pos].arg1);

      // NOTE: cmd_pos is not incremented, hence this information will
      // be overwritten; odd, but this is how the original ipfw
      // software works as well; we suspect that some default values
      // make it work anyway
    }
  else
    {
      // normal IP address with optional port separated by ':'
      if (atoaddr (src, &addr, &src_port) < 0)
	{
	  WARNING ("Invalid IP address: %s", src);
	  return ERROR;
	}

      cmd[cmd_pos].opcode = O_IP_SRC;
      cmd[cmd_pos].len = 2;
      cmd[cmd_pos].arg1 = 0;
      ((uint32_t *) cmd)[cmd_pos + 1] = addr.word;
      DEBUG ("Cmd src IP: opcode=%d len=%d arg1=%d",
	     cmd[cmd_pos].opcode, cmd[cmd_pos].len, cmd[cmd_pos].arg1);
      DEBUG ("Cmd src IP+1: %u", ((uint32_t *)cmd)[cmd_pos + 1]);

      cmd_pos += 2;
    }

  // process destination address
  if (strncmp (dst, "any", strlen (dst)) == 0)
    {
      // destination address was "any"
      cmd[cmd_pos].opcode = O_IP_DST;
      cmd[cmd_pos].len = 0;
      cmd[cmd_pos].arg1 = 0;
      DEBUG ("Cmd dst any: opcode=%d len=%d arg1=%d",
	     cmd[cmd_pos].opcode, cmd[cmd_pos].len, cmd[cmd_pos].arg1);

      // NOTE: cmd_pos is not incremented, hence this information will
      // be overwritten; odd, but this is how the original ipfw
      // software works as well; we suspect that some default values
      // make it work anyway
    }
  else
    {
      // normal IP address with optional port separated by ':'
      if (atoaddr (dst, &addr, &dst_port) < 0)
	{
	  WARNING ("Invalid IP address: %s", dst);
	  return ERROR;
	}

      cmd[cmd_pos].opcode = O_IP_DST;
      cmd[cmd_pos].len = 2;
      cmd[cmd_pos].arg1 = 0;
      ((uint32_t *) cmd)[cmd_pos + 1] = addr.word;
      DEBUG ("Cmd dst IP: opcode=%d len=%d arg1=%d",
	     cmd[cmd_pos].opcode, cmd[cmd_pos].len, cmd[cmd_pos].arg1);
      DEBUG ("Cmd dst IP+1: %u", ((uint32_t *)cmd)[cmd_pos + 1]);

      cmd_pos += 2;
    }

  // process source port if provided
  if (strncmp (src, "any", strlen (src)) != 0)
    {
      if (src_port > 0)
	{
	  cmd[cmd_pos].opcode = O_IP_SRCPORT;
	  cmd[cmd_pos].len = 2;
	  cmd[cmd_pos].arg1 = 0;
	  ((uint32_t *) cmd)[cmd_pos + 1] = src_port | src_port << 16;
	  cmd_pos += 2;
	}
    }

  // process destination port if provided
  if (strncmp (dst, "any", strlen (dst)) != 0)
    {
      if (dst_port > 0)
	{
	  cmd[cmd_pos].opcode = O_IP_DSTPORT;
	  cmd[cmd_pos].len = 2;
	  cmd[cmd_pos].arg1 = 0;
	  ((uint32_t *) cmd)[cmd_pos + 1] = dst_port | dst_port << 16;
	  cmd_pos += 2;
	}
    }

  // use in/out direction indicators; 
  // if both directions are requested, nothing needs to be done
  if (direction != DIRECTION_BOTH)
    {
      // basic command code for in/out operation
      cmd[cmd_pos].opcode = O_IN;
      cmd[cmd_pos].len = 1;
      cmd[cmd_pos].arg1 = 0;
      // a negation mask is used for meaning "out"
      if (direction == DIRECTION_OUT)
	cmd[cmd_pos].len |= F_NOT;

      cmd_pos += 1;
    }

  // configure pipe
  cmd[cmd_pos].opcode = O_PIPE;
  cmd[cmd_pos].len = 1;
  cmd[cmd_pos].arg1 = pipe_nr;
  ((uint32_t *) cmd)[cmd_pos + 1] = 0;
  cmd_pos += 1;	// trick!

  // set size of rule structure;
  // NOTE: cmd_pos must be decremented by 1; we assume it is because
  // the first command is already taken into account by the ip_fw structure
  rule_len = sizeof (struct ip_fw) + (cmd_pos-1) * 4;

  // allocate rule and set its content to zero
  if ((rule = (struct ip_fw *) malloc (sizeof (struct ip_fw)
				       + cmd_pos * 4)) == NULL)
    {
      WARNING ("Could not allocate memory for a new rule");
      return ERROR;
    }
  bzero (rule, sizeof (struct ip_fw));

  // copy all commands to the rule structure
  bcopy (cmd, &rule->cmd, cmd_pos * 4);

  // store length of the sequence of commands as the last position in the array
  rule->cmd_len = cmd_pos;

  // store the offset of actions as a pointer to the last command
  rule->act_ofs = cmd_pos - 1;

  // store the rule number from the function argument
  rule->rulenum = rulenum;

#ifdef MESSAGE_DEBUG
  DEBUG ("Add ipfw rule '%d' and pipe '%d' from '%s' to '%s' (direction %s) \
(rule_len=%d): ", rulenum, pipe_nr, src, dst, 
	 (direction==DIRECTION_OUT)?"OUT":
	 ((direction==DIRECTION_IN)?"IN":"BOTH"), rule_len);
  //print_rule (rule);
#endif

  // do socket operation
  if (getsockopt (socket_id, IPPROTO_IP, IP_FW_ADD, rule, &rule_len) < 0)
    {
      WARNING ("Error getting socket options.");
      perror ("getsockopt");
      WARNING ("Adding rule & pipe operation failed.");
      return ERROR;
    }

  free(rule);

  return SUCCESS;
}

// get ipfw rules; on return, rule data will be stored in 'rule_data';
// if 'rule_data' must be allocated, set 'rule_data_alloc_size' to
// UNDEFINED_RULE_DATA_SIZE when calling, and on return it wil contain
// the allocated size, otherwise it is assumed 'rule_data_alloc_size'
// contains the allocate size of 'rule_data' (this is useful for next
// call, to avoid reallocation each time); on return 'rules' is an array of
// pointers to consecutive rules, and 'rules_count' is the number of rules;
// NOTE: when calling rule_count should contain the maximum number of rules
// that can be used;
// return SUCCESS on success, ERROR on error
int
get_rules (uint socket_id, struct ip_fw **rule_data,
	   int *rule_data_alloc_size, struct ip_fw **rules, int *rule_count)
{
  struct ip_fw *rule_pointer;
  int MAX_RULE_COUNT = *rule_count;

#define NEXT_RULE(rule_pointer)	((struct ip_fw *)((char *)rule_pointer       \
						  + RULESIZE(rule_pointer)))

  // number of bytes to be allocated
  int n_alloc;

  // number of bytes passed to get_socket_options
  int n_bytes;

  // pointer limiting the allocated block 'rule_data'  
  char *limit_pointer;

  int read_success = FALSE;


  // check whether buffer needs to be allocated
  if (*rule_data_alloc_size == UNDEFINED_ALLOC_SIZE)
    {
      DEBUG ("Buffer needs to be allocated");
      // initialize 'n_alloc' to an arbitrary value 
      n_alloc = 1024;

      // try to reallocate
      if (((*rule_data) = realloc (*rule_data, n_alloc)) == NULL)
	{
	  WARNING ("realloc failed");
	  return ERROR;
	}
    }
  else
    {
      DEBUG ("Buffer is already allocated");
      n_alloc = *rule_data_alloc_size;
    }

  // initialize 'n_bytes'
  n_bytes = n_alloc;

  // repeat socket operation until success
  do
    {
      DEBUG ("Get ipfw rules.");
      // do socket operation
      if (getsockopt (socket_id, IPPROTO_IP, IP_FW_GET, *rule_data,
			      (socklen_t *) & n_bytes) < 0)
	{
	  WARNING ("Error getting socket options");
	  perror ("getsockopt");
	  WARNING ("Getting rule operation failed");
	  return ERROR;
	}

      // check whether the answer to be returned is larger than
      // the allocated block
      if (n_bytes >= n_alloc)
	{
	  // need to reallocate => increase the number of bytes to
	  // be allocated (arbitrary values)
	  n_alloc = n_alloc * 2 + 20;
	  n_bytes = n_alloc;
	  // try to reallocate
	  if (((*rule_data) = realloc (*rule_data, n_bytes)) == NULL)
	    {
	      WARNING ("realloc failed");
	      return ERROR;
	    }
	  else
	    *rule_data_alloc_size = n_alloc;
	}
      else
	read_success = TRUE;
    }
  while (read_success == FALSE);

  // save last allocated block size
  *rule_data_alloc_size = n_alloc;


  // go through rules; they have variable size so we
  // need to scan the list

  // NOTE: the last rule always has the number 65535, and cannot be
  // modified or deleted

  // initialize number of rules
  *rule_count = 0;
  limit_pointer = (char *) (*rule_data) + n_bytes;

  // find start for each rule and count them
  for (rule_pointer = (*rule_data); (char *) rule_pointer < limit_pointer;
       rule_pointer = NEXT_RULE (rule_pointer))
    {
      DEBUG ("Rule '%d': %"PRIu64" pkts, %"PRIu64" bytes.",
	     rule_pointer->rulenum, align_uint64 (&rule_pointer->pcnt), 
	     align_uint64 (&rule_pointer->bcnt));

      if ((*rule_count) >= MAX_RULE_COUNT)
	{
	  WARNING ("Rule count exceeds maximum allowed count (%d)",
		   MAX_RULE_COUNT);
	  return ERROR;
	}
      else
	{
	  rules[*rule_count] = rule_pointer;
	  (*rule_count)++;
	}
    }

  return SUCCESS;
}

// delete an ipfw rule specified by 'rule_number';
// return SUCCESS on success, ERROR on error
int
delete_rule (uint socket_id, u_int32_t rule_number)
{
  // Note: rule number is of type u_int16_t in ip_fw.h,
  // but a comment in ip_fw2.c shows that the expected size
  // when applying socket options is u_int32_t (see comment below)

  /* COMMENT FROM IP_FW2.C
   *
   * IP_FW_DEL is used for deleting single rules or sets,
   * and (ab)used to atomically manipulate sets. Argument size
   * is used to distinguish between the two:
   *    sizeof(u_int32_t)
   *    delete single rule or set of rules,
   *    or reassign rules (or sets) to a different set.
   *    2*sizeof(u_int32_t)
   *    atomic disable/enable sets.
   *    first u_int32_t contains sets to be disabled,
   *    second u_int32_t contains sets to be enabled.
   */

  DEBUG ("Deleting ipfw rule '%d'...", rule_number);
  if (setsockopt (socket_id, IPPROTO_IP, IP_FW_DEL, 
		  &rule_number, sizeof (rule_number)) < 0)
    {
      WARNING ("Error setting socket options.");
      perror ("setsockopt");
      WARNING ("Delete rule operation failed.");
      return ERROR;
    }
  DEBUG ("Deleted ipfw rule '%d'.", rule_number);

  return SUCCESS;
}

// print an instruction given in 'insn'
void
print_insn (ipfw_insn *insn, unsigned cmd_len)
{
  int i;
  INFO ("\tInstruction content (size=%u):", cmd_len);
  for (i=0; i<cmd_len; i++)
    INFO ("Cmds[%d]: opcode=%u len=%u arg1=%u", i,
	   insn[i].opcode, insn[i].len, insn[i].arg1);
}

// print a rule structure given in 'rule'
void
print_rule (struct ip_fw *rule)
{
  int i;
  uint8_t *byte_array;
  unsigned rule_length;

  INFO ("Rule content (size=%zu bytes):", sizeof (*rule));
  INFO ("\tx_next=%p next_rule=%p", rule->x_next, rule->next_rule);
  INFO ("\tact_ofs=%u cmd_len=%u rulenum=%u set=%u _pad=%u id=%u",
	  rule->act_ofs, rule->cmd_len, rule->rulenum, rule->set, 
	  rule->_pad, rule->id);
  INFO ("\tpcnt=%"PRIu64" bcnt=%"PRIu64" timestamp=%u", 
	  rule->pcnt, rule->bcnt, rule->timestamp);
  print_insn (rule->cmd, rule->cmd_len);

  byte_array = (uint8_t *) rule;
  rule_length = sizeof (struct ip_fw) + (rule->cmd_len-1) * 4;
  INFO ("Rule content in binary form (rule_length=%d bytes):",
	 rule_length);
  for (i=0; i<rule_length; i++)
    printf("%d ", byte_array[i]);
  printf("\n");
}


////////////////////////////////////////////////
// dummynet pipe related functions
////////////////////////////////////////////////

// configure a dummynet pipe identified by 'pipe_nr';
// the parameters to use are 'bandwidth' [bps], 'delay' [ms],
// 'loss_rate' [probability], and 'queue_size' [packets];
// return SUCCESS on success, ERROR on error
int
configure_pipe (int socket_id, int pipe_nr, float bandwidth,
		float delay, float loss_rate, int queue_size)
{

  if (bandwidth < 0 || bandwidth > MAX_BANDWIDTH || !isfinite(bandwidth))
    {
      WARNING ("Bandwidth value %f is not within the allowed range [%.2f,%.2f]. Skipping pipe configuration.", bandwidth, 0.0, MAX_BANDWIDTH);
      return ERROR;
    }

  if (delay < 0 || delay > MAX_DELAY || !isfinite(delay))
    {
      WARNING ("Delay value %f is not within the allowed range [%.2f,%.2f]. Skipping pipe configuration.", delay, 0.0, MAX_DELAY);
      return ERROR;
    }

  if (loss_rate < 0 || loss_rate > MAX_LOSS_RATE || !isfinite(loss_rate))
    {
      WARNING ("Loss rate value %f is not within the allowed range [%.2f,%.2f]. Skipping pipe configuration.", loss_rate, 0.0, MAX_LOSS_RATE);
      return ERROR;
    }

  // lines below are based on code in dummynet.c

  // pipe structure
  uint8_t cmd_buf[sizeof(struct dn_id) 
		  + sizeof(struct dn_sch) + sizeof(struct dn_link) 
		  + sizeof(struct dn_fs)];
  int cmd_buf_size;

  struct dn_id   *header;
  struct dn_sch  *sch;
  struct dn_link *link;
  struct dn_fs   *fs;

  // header + two structures (same as above!)
  cmd_buf_size = sizeof(struct dn_id) 
    + sizeof(struct dn_sch) + sizeof(struct dn_link) 
    + sizeof(struct dn_fs);

  header = (struct dn_id *)&(cmd_buf[0]);
  sch = (struct dn_sch *)&(cmd_buf[sizeof(struct dn_id)]);
  link = (struct dn_link*)
    &(cmd_buf[sizeof(struct dn_id) + sizeof(struct dn_sch)]);
  fs = (struct dn_fs *)
    &(cmd_buf[sizeof(struct dn_id) 
	      + sizeof(struct dn_sch) + sizeof(struct dn_link)]);
  
  // reset data structures
  bzero (header, sizeof (*header));
  bzero (sch, sizeof (*sch));
  bzero (link, sizeof (*link));
  bzero (fs, sizeof (*fs));

  // header initialization
  header->len = sizeof(struct dn_id);
  header->type = DN_CMD_CONFIG;
  header->subtype = 0;
  header->id = DN_API_VERSION;

  // scheduler initialization
  sch->oid.len = sizeof(struct dn_sch);
  sch->oid.type = DN_SCH;
  sch->oid.subtype = 0;
  sch->oid.id = 0;
  sch->flags |= DN_PIPE_CMD;

  // link initialization
  link->oid.len = sizeof(struct dn_link);
  link->oid.type = DN_LINK;
  link->oid.subtype = 0;
  link->oid.id = 0;

  // fs initialization
  fs->oid.len = sizeof(struct dn_fs);
  fs->oid.type = DN_FS;
  fs->oid.subtype = 0;
  fs->oid.id = 0;


  // set pipe number to scheduler, link, and fs
  sch->sched_nr = pipe_nr;
  link->link_nr = pipe_nr;
  fs->fs_nr = pipe_nr + 2*DN_MAX_ID;
  fs->sched_nr = pipe_nr + DN_MAX_ID;

  // prepare adjusted values for parameters

  // 1. if packet size is known bandwidth could be adjusted:
  //   multiplication factor 1.0778 was computed for 400 byte 
  //   datagrams because of 28 bytes header (428/400=1.07)
  //   however we consider bandwidth at Ethernet level, 
  //   therefore we don't multiply here but when plotting results

  //   Note: ip_dummynet.h says that bandwidth is either in bytes/tick
  //         or in bits/second; we use the latter
  link->bandwidth = (int) round (bandwidth);	// * 2.56);

  // 2. no adjustment necessary for delay expressed in ms
  link->delay = (int) round (delay);

  // set burst to 0
  link->burst = 0;

  // 3. loss rate probability must be extended to 
  //    2^31-1 (=0x7fffffff=2147483647) is equivalent to 100% loss
  double plr_double = loss_rate * 0x7fffffff;
  int32_t plr_int32 = 
    (plr_double > 2147483647)?2147483647:(int32_t)floor(plr_double);
  //fs->plr = (int) (loss_rate * 0x7ffffffe);
  fs->plr = plr_int32;

  // you may set the pipe queue size to a small value to avoid large
  // delays in case bandwidth limitation is enforced
  fs->qsize = queue_size;

  // do socket operation
#ifdef MESSAGE_DEBUG
  DEBUG ("Configure ipfw pipe '%d':", pipe_nr);
  //print_pipe (link, fs);

  //Debugging code by Razvan added on 2011-06-24
  //{
  //  int i;
  //  uint8_t *byte_array;
    
  //  DEBUG ("Command in binary form (cmd_size=%u):", cmd_buf_size);
  //  byte_array = (uint8_t *) cmd_buf;
  //  for (i=0; i<cmd_buf_size; i++)
  //    printf("%d ", byte_array[i]);
  //  printf("\n");
  //}
#endif

  // do socket operation
  if (setsockopt (socket_id, IPPROTO_IP, IP_DUMMYNET3,
		  cmd_buf, cmd_buf_size) < 0)
    {
      WARNING ("Error setting socket options");
      perror ("setsockopt");
      WARNING ("Pipe configuration could not be applied");
      return ERROR;
    }

  return SUCCESS;
}

// delete an ipfw pipe identified by 'pipe_number' by using a system call;
// return SUCCESS on success, ERROR on error
/*
int
delete_pipe_syscall (int pipe_number)
{
  // this is an example of using system calls for controlling ipfw;
  // however, we prefer socket operations since they are supposed
  // to be more efficient

  char cmd[MAX_STRING];
  int status;

  snprintf (cmd, MAX_STRING - 1, "ipfw pipe %d delete", pipe_number);

  INFO ("Making system call: %s", cmd);

  status = system (cmd);

  if (status != 0)
    return ERROR;
  else
    return SUCCESS;
}
*/

// delete an ipfw pipe identified by 'pipe_number' by using socket operations;
// return SUCCESS on success, ERROR on error
int
delete_pipe (uint socket_id, int pipe_number)
{
  // the code below is based on code from dummynet.c and ip_dn_glue.c
  // in the ipfw3 release

  struct
  {
    struct dn_id oid;
    uintptr_t a[1];	// add more if lists are to be supported
  } cmd;

  int do_pipe = 1; // it is a pipe

  // fill in structure fields
  cmd.oid.len = sizeof(cmd);
  cmd.oid.type = DN_CMD_DELETE;
  cmd.oid.id = DN_API_VERSION;
  cmd.oid.subtype = (do_pipe == 1) ? DN_LINK :
    ( (do_pipe == 2) ? DN_FS : DN_SCH);
  cmd.a[0] = pipe_number;

  DEBUG ("Deleting ipfw pipe '%d'...", pipe_number);

  // do socket operation
  if (setsockopt (socket_id, IPPROTO_IP, IP_DUMMYNET3, &cmd, cmd.oid.len) < 0)
    {
      WARNING ("Error setting socket options.");
      perror ("setsockopt");
      WARNING ("Delete pipe operation failed.");
      return ERROR;
    }
  DEBUG ("Deleted ipfw pipe '%d'.", pipe_number);

  return SUCCESS;
}

// print a pipe structure given by 'pipe'
void
print_pipe (struct dn_link *link, struct dn_fs *scheduler)
{
  INFO ("Pipe content (size=%zu):", sizeof (*link));
  INFO ("\tLINK: link_nr=%u bandwidth=%u delay=%u burst=%"PRIu64"", 
	 link->link_nr, link->bandwidth, link->delay, link->burst);
  INFO ("\tSCHEDULER: fs_nr=%u qsize=%u plr=%u", 
	 scheduler->fs_nr, scheduler->qsize, scheduler->plr);
 /* old code
#ifdef NEW_DUMMYNET
  INFO ("\tbandwidth=%u delay=%u delay_type=%u fs.plr=%u fs.qsize=%u",
	  pipe->bandwidth, pipe->delay, pipe->delay_type, pipe->fs.plr,
	  pipe->fs.qsize);
#endif
  */
}

