
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 *
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: management.c
 * Function: Management functions of the wireconf library
 *
 * Authors: Junya Nakata, Razvan Beuran
 *
 ***********************************************************************/

#include <sys/queue.h>
#include <sys/socket.h>

#include <net/if.h>
//#include <netinet/in.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <inttypes.h>

#include <time.h>

#include "wireconf.h"

////////////////////////////////////////////////
// wireconf library management functions
////////////////////////////////////////////////

// initialize the wireconf library; static data is stored in the
// structure 'wireconf'; 
// return SUCCESS on succes, ERROR on error
int
wireconf_init (struct wireconf_class *wireconf, int my_id,
	       in_addr_t * IP_addresses, int node_count)
{
  int i;

  // get starting time
  clock_gettime (CLOCK_REALTIME, &(wireconf->start_time));

  // store id and node count  
  wireconf->my_id = my_id;
  wireconf->node_count = node_count;

  // allocate array for storing IP addresses
  wireconf->IP_addresses =
    malloc (wireconf->node_count * sizeof (struct in_addr));

  if (wireconf->IP_addresses == NULL)
    {
      WARNING ("Cannot allocate memory for IP_addresses.");
      return ERROR;
    }

  // copy IP address information
  for (i = 0; i < wireconf->node_count; i++)
    (wireconf->IP_addresses[i]).s_addr = IP_addresses[i];


  // allocate array for storing total channel utilizations
  wireconf->total_channel_utilizations =
    malloc (wireconf->node_count * sizeof (float));
  if (wireconf->total_channel_utilizations == NULL)
    {
      WARNING ("Cannot allocate memory for total_channel_utilizations.");
      return ERROR;
    }

  // allocate array for storing total transmission probabilities
  wireconf->total_transmission_probabilities =
    malloc (wireconf->node_count * sizeof (float));
  if (wireconf->total_transmission_probabilities == NULL)
    {
      WARNING ("Cannot allocate memory for total_transmission_probabilities.");
      return ERROR;
    }


  // Lan allocate array for storing statistic of channel utilization per source
  wireconf->channel_utilization_per_source =
    malloc (wireconf->node_count * sizeof (float *));
  if (wireconf->channel_utilization_per_source == NULL)
    {
      WARNING ("Cannot allocate memory for channel_utilization_per_source.");
      return ERROR;
    }
  for (i = 0; i < wireconf->node_count; i++)
    {
      wireconf->channel_utilization_per_source[i] =
	malloc (wireconf->node_count * sizeof (float));
      if (wireconf->channel_utilization_per_source[i] == NULL)
	{
	  WARNING ("Cannot allocate memory for channel_utilization_per_source[i].");
	  return ERROR;
	}     
    }


  // allocate array for storing self channel utilizations
  wireconf->self_channel_utilizations =
    malloc (wireconf->node_count * sizeof (float));

  if (wireconf->self_channel_utilizations == NULL)
    {
      WARNING ("Cannot allocate memory for self_channel_utilizations.");
      return ERROR;
    }

  // allocate array for storing self transmission probabilities
  wireconf->self_transmission_probabilities =
    malloc (wireconf->node_count * sizeof (float));

  if (wireconf->self_transmission_probabilities == NULL)
    {
      WARNING ("Cannot allocate memory for self_transmission_probabilities.");
      return ERROR;
    }

  // initialize channel_utilizations
  for (i = 0; i < wireconf->node_count; i++)
    {
      wireconf->total_channel_utilizations[i] = 0.0;
      wireconf->total_transmission_probabilities[i] = 0.0;
      wireconf->self_channel_utilizations[i] = 0.0;
      wireconf->self_transmission_probabilities[i] = 0.0;
    }

  // start wireconf library message
  INFO ("\nWireconf library for my_id=%d (node_count=%d) my_IP=%s started on \
%s", wireconf->my_id, wireconf->node_count, 
	inet_ntoa (wireconf->IP_addresses[wireconf->my_id]), 
	ctime (&(wireconf->start_time.tv_sec)));

  // initialize do_interrupt flags
  wireconf->do_interrupt_send = FALSE;
  wireconf->do_interrupt_listen = FALSE;

  // start statistics listen thread
  DEBUG ("Starting statistics listen thread...");
  if (pthread_create (&(wireconf->stats_listen_thread), NULL,
		      stats_listen_thread, (void *) wireconf) != 0)
    {
      WARNING ("Cannot create statistics listen thread.");
      return ERROR;
    }

  // start statistics send thread
  DEBUG ("Starting statistics send thread...");
  if (pthread_create (&(wireconf->stats_send_thread), NULL,
		      stats_send_thread, (void *) wireconf) != 0)
    {
      WARNING ("Cannot create statistics send thread.");
      return ERROR;
    }

  // create pipe used to provide statistics data to the send thread
  if (pipe (wireconf->stats_provide_pipe) == -1)
    {
      WARNING ("Cannot create pipe of statistics send thread.");
      perror ("pipe");
      return ERROR;
    }

  return SUCCESS;
}

// finalize wireconf execution and wait for end of thread execution;
// return SUCCESS on succes, ERROR on error
int
wireconf_finalize (struct wireconf_class *wireconf)
{
  int i;

  // free array for storing IP addresses
  if (wireconf->IP_addresses != NULL)
    free (wireconf->IP_addresses);

  // free array of total channel utilizations
  if (wireconf->total_channel_utilizations != NULL)
    free (wireconf->total_channel_utilizations);

  // free array of total transmission probabilities
  if (wireconf->total_transmission_probabilities != NULL)
    free (wireconf->total_transmission_probabilities);

  // free array of channel utilization per source
  for (i = 0; i < wireconf->node_count; i++)
    free(wireconf->channel_utilization_per_source[i]);
  free(wireconf->channel_utilization_per_source);

  // free array of self channel utilizations
  if (wireconf->self_channel_utilizations != NULL)
    free (wireconf->self_channel_utilizations);

  // free array of self transmission probabilities
  if (wireconf->self_transmission_probabilities != NULL)
    free (wireconf->self_transmission_probabilities);

  ////////////////////////////////////////////////////
  // stopping statistics send thread
  DEBUG ("Stopping wireconf statistics send thread...");
  wireconf->do_interrupt_send = TRUE;

  // close write end of the pipe
  // (this will make the reading end finish execution)
  close (wireconf->stats_provide_pipe[1]);

  // wait for thread to react to flag and pipe changes
  sleep (1);

  // close read end of the pipe
  close (wireconf->stats_provide_pipe[0]);

  // stop and wait for statistics send thread to end
  pthread_cancel (wireconf->stats_send_thread);
  if (pthread_join (wireconf->stats_send_thread, NULL) != 0)
    {
      WARNING ("Error joining statistics send thread");
      return ERROR;
    }


  ////////////////////////////////////////////////////
  // stopping statistics listen thread
  DEBUG ("Stopping wireconf statistics listen thread...");
  wireconf->do_interrupt_listen = TRUE;

  // wait for thread to react to flag change
  sleep (1);

  // stop and wait for statistics listen thread to end
  pthread_cancel (wireconf->stats_listen_thread);
  if (pthread_join (wireconf->stats_listen_thread, NULL) != 0)
    {
      WARNING ("Error joining statistics listen thread");
      return ERROR;
    }

  return SUCCESS;
}

// deliver traffic statistics to wireconf library for distribution
// to other instances;
// return SUCCESS on succes, ERROR on error
int
wireconf_deliver_stats (struct wireconf_class *wireconf,
			struct stats_class *stats)
{
  if (stats->channel_utilization >= 1.0)
    stats->channel_utilization = 0.9999;

  if (stats->transmission_probability >= 1.0)
    stats->transmission_probability = 0.9999;

  DEBUG ("Deliver statistics to wireconf send thread:		\
channel_utilization=%.4f transmission_probability=%.4f...", 
  	stats->channel_utilization, stats->transmission_probability);

  // store data in local structure
  wireconf->total_channel_utilizations[wireconf->my_id] =
    stats->channel_utilization;
  wireconf->total_transmission_probabilities[wireconf->my_id] =
    stats->transmission_probability;

  // send data to pipe (the statistics send thread is reading at the
  // other end
  if (write (wireconf->stats_provide_pipe[1], stats,
	     sizeof (struct stats_class)) == -1)
    {
      WARNING ("Error writing to pipe");
      perror ("write");
      return ERROR;
    }

  return SUCCESS;
}

// convert a "struct timespec" time value to seconds
static __inline float
timespec2sec (struct timespec *time_spec)
{
  return ((float) time_spec->tv_sec + (float) time_spec->tv_nsec / 1e9);
}

// compute the difference in seconds between two "struct timespec" time values
static __inline float
timespec_diff2sec (struct timespec *time_spec1, struct timespec *time_spec2)
{
  /*
     return ((float)(time_spec1->tv_sec - time_spec2->tv_sec) + 
     (float)(time_spec1->tv_nsec - time_spec2->tv_nsec) / 1e9);
   */

  // the algorithm below is inspired from: 
  // http://www.delorie.com/gnu/docs/glibc/libc_428.html
  // but it is not yet clear if it is better than the above simple calculation
  int sec_no;
  struct timespec result;

  // carry for the later subtraction by updating time_spec2
  if (time_spec1->tv_nsec < time_spec2->tv_nsec)
    {
      sec_no = (time_spec2->tv_nsec - time_spec1->tv_nsec) / 1e9 + 1;
      time_spec2->tv_nsec -= (1e9 * sec_no);
      time_spec2->tv_sec += sec_no;
    }

  if (time_spec1->tv_nsec - time_spec2->tv_nsec > 1e9)
    {
      sec_no = (time_spec1->tv_nsec - time_spec2->tv_nsec) / 1e9;
      time_spec2->tv_nsec += (1e9 * sec_no);
      time_spec2->tv_sec -= sec_no;
    }

  // compute the time difference; tv_nsec is certainly positive
  result.tv_sec = time_spec1->tv_sec - time_spec2->tv_sec;
  result.tv_nsec = time_spec1->tv_nsec - time_spec2->tv_nsec;

  return timespec2sec (&result);
}

// get the time elapsed (in seconds) since wireconf started execution
float
get_elapsed_time (struct wireconf_class *wireconf)
{
  struct timespec crt_time;

  // get current time
  clock_gettime (CLOCK_REALTIME, &crt_time);

  return timespec_diff2sec (&crt_time, &(wireconf->start_time));
}
