
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 *
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: routing_info.c
 * Function: Get routing information from kernel routing table and map it 
 *           to and from internal QOMET node id
 *
 * Authors: Lan Nguyen Tien, Razvan Beuran
 *
 *   $Revision: 128 $
 *   $LastChangedDate: 2009-02-06 10:21:50 +0900 (Fri, 06 Feb 2009) $
 *   $LastChangedBy: razvan $
 *
 ***********************************************************************/

#include "wireconf.h"

///////////////////////////////////
// Routing table related functions
///////////////////////////////////

// given the arrays of IP addresses in binary and char format
// 'IP_addresses' and 'IP_char_addresses', get the next hop id
// for the route to destination 'dst_id'; 'direction' can be either
// DIRECTION_IN (incoming route) or DIRECTION_OUT (outgoing route);
// return the next_hop_id on success, ERROR on error
int
get_next_hop_id (in_addr_t * IP_addresses, char *IP_char_addresses,
		 int dst_id, int direction)
{
  char dst_ip[IP_ADDR_SIZE], next_hop_ip[IP_ADDR_SIZE];
  int next_hop_id;

  // is DIRECTION_IN next hop meaningful?!; equivalent to previous hop?!

  // for incoming routes, the route is the destination itself
  if (direction == DIRECTION_IN)
    return dst_id;

  // convert destination id to IP address
  if (get_ip_from_id (IP_char_addresses, dst_id, dst_ip) == ERROR)
    return ERROR;

#ifdef __FreeBSD__
  // get the next hop IP address for a destination IP address
  if (get_next_hop_ip_freebsd (dst_ip, next_hop_ip) == ERROR)
    return ERROR;
#elif __linux__
  // get the next hop IP address for a destination IP address
  if (get_next_hop_ip_linux (dst_ip, next_hop_ip) == ERROR)
    return ERROR;
#endif

  // get the next hop id from the next hop IP address
  if ((next_hop_id = get_id_from_ip_string (IP_addresses, next_hop_ip))
      != ERROR)
    return next_hop_id;

  return ERROR;
}

#ifdef __FreeBSD__
// get from the routing table the next hop associated to destination with
// IP address 'dst'; the result IP address is stored in the argument 
// 'next_hop';
// return SUCCESS on success, ERROR on error
int
get_next_hop_ip_freebsd (char *dst, char *next_hop)
{
  int sockfd;
  char *buf;
  pid_t pid;

  struct rt_msghdr *rtm;
  struct sockaddr *sa, *rti_info[RTAX_MAX];
  struct sockaddr_in *sin;

  ssize_t n;
  char *gw;
  char *rt_dst;

  buf = calloc (1, ROUTE_BUFFER_SIZE);
  rtm = (struct rt_msghdr *) buf;
  sin = (struct sockaddr_in *) (rtm + 1);

  sin->sin_family = AF_INET;
  sin->sin_len = sizeof (struct sockaddr_in);
  inet_pton (AF_INET, dst, &(sin->sin_addr));

  sockfd = socket (PF_ROUTE, SOCK_RAW, 0);

  if (sockfd == -1)
    {
      WARNING ("Cannot open socket to routing table");
      exit (1);
    }

  rtm = (struct rt_msghdr *) buf;
  rtm->rtm_msglen = sizeof (struct rt_msghdr) + sizeof (struct sockaddr_in);
  rtm->rtm_version = RTM_VERSION;
  rtm->rtm_type = RTM_GET;
  rtm->rtm_addrs = RTA_DST;
  rtm->rtm_pid = pid = getpid ();
  rtm->rtm_seq = RTM_SEQ_CT;

  // send request through routing socket
  write (sockfd, rtm, rtm->rtm_msglen);

  // receive the return value from the socket
  do
    {
      n = read (sockfd, rtm, ROUTE_BUFFER_SIZE);
    }
  while (rtm->rtm_type != RTM_GET || rtm->rtm_seq != RTM_SEQ_CT ||
	 rtm->rtm_pid != pid);

  rtm = (struct rt_msghdr *) buf;
  sa = (struct sockaddr *) (rtm + 1);

  // read from sockaddr sa & put result in to rti_info
  get_routing_table_addr (rtm->rtm_addrs, sa, rti_info);

  // free buffer
  free (buf);

  // return gateway IP address
  if ((sa = rti_info[RTAX_DST]) != NULL)
    rt_dst = sock_ntop_host (sa, sa->sa_len);
  else
    return ERROR;

  // if rt_dst != NULL and rt_dst equals dst then gw is considered

  //if (!strcmp(rt_dst, dst)) {
  if (1)
    {
      if ((sa = rti_info[RTAX_GATEWAY]) != NULL)
	{
	  gw = (char *) sock_ntop_host (sa, sa->sa_len);

	  // if gw contains AF_LINK then dst can be reached in 1 hop
	  if (strncmp (gw, "AF_LINK", 7))
	    {
	      strncpy (next_hop, gw, 20);
	      close (sockfd);
	      return SUCCESS;
	    }
	  else
	    {
	      strncpy (next_hop, dst, 20);
	      close (sockfd);
	      return SUCCESS;
	    }
	}
      else
	{
	  close (sockfd);
	  return ERROR;
	}
    }
  else
    {
      close (sockfd);
      return ERROR;
    }

  close (sockfd);
}
#endif

// get from the routing table the next hop associated to destination with
// IP address 'dst'; the result IP address is stored in the argument 
// 'next_hop';
// return SUCCESS on success, ERROR on error
/*
int get_next_hop_ip(char *dst, char *next_hop)
{
  int sockfd;
  char *buf;
  pid_t pid;

  struct sockaddr *sa; 
  struct sockaddr *rti_info[RTAX_MAX];
  struct sockaddr_in *sin;
  char *gw;

  int status = SUCCESS;

  DEBUG ("Getting next hop IP address for destination '%s'...", dst);

#ifdef __FreeBSD__
  struct rt_msghdr *rtm;
#elif __linux__
  memset(rti_info, 0, sizeof(rti_info));
  struct rtmsg *rtm;
#endif

#ifdef __FreeBSD__
  ssize_t n;
  char *rt_dst;

  buf = calloc(1, ROUTE_BUFFER_SIZE);
  rtm = (struct rt_msghdr *) buf;
  sin = (struct sockaddr_in *) (rtm + 1);
#elif __linux__
  buf = calloc(1, ROUTE_BUFFER_SIZE);
  rtm = (struct rtmsg *)buf;
  sin = (struct sockaddr_in *) (rtm + 1);
#endif

  sin->sin_family = AF_INET;
#ifdef __FreeBSD__
  sin->sin_len = sizeof(struct sockaddr_in);
#endif
  inet_pton(AF_INET, dst, &(sin->sin_addr));

  sockfd = socket(PF_ROUTE, SOCK_RAW, 0);

  if(sockfd == -1)
    {
      WARNING("Cannot open socket to routing table");
      return ERROR;
    }

#ifdef __FreeBSD__
  rtm = (struct rt_msghdr *) buf;
  rtm->rtm_msglen = sizeof(struct rt_msghdr) + sizeof(struct sockaddr_in);
  rtm->rtm_version = RTM_VERSION;
  rtm->rtm_type = RTM_GET;
  rtm->rtm_addrs = RTA_DST;
  rtm->rtm_pid = pid = getpid();
  rtm->rtm_seq = RTM_SEQ_CT;

  // send request through routing socket
  write(sockfd, rtm, rtm->rtm_msglen);

  // receive the return value from the socket
  do
    {
      n = read(sockfd, rtm, ROUTE_BUFFER_SIZE);
    } 
  while(rtm->rtm_type != RTM_GET || rtm->rtm_seq != RTM_SEQ_CT || 
	rtm->rtm_pid != pid);

  rtm = (struct rt_msghdr *) buf;
  sa = (struct sockaddr *) (rtm + 1);

  // read from sockaddr sa & put result in to rti_info
  get_routing_table_addr(rtm->rtm_addrs, sa, rti_info);

  // free buffer
  free(buf);
  
  // return gateway IP address
  if((sa = rti_info[RTAX_DST]) != NULL)
    rt_dst = sock_ntop_host(sa, sa->sa_family);
  else
    {
      WARNING ("Error in rti_info");
      return ERROR;
    }
#elif __linux__
  rtm = (struct rtmsg *)buf;
  rtm->rtm_type = RTM_GETROUTE;
  pid = getpid();
#endif
  // if rt_dst != NULL and rt_dst equals dst then gw is considered

  //if (!strcmp(rt_dst, dst)) {
  if(1)
    {
#ifdef __FreeBSD__
      if((sa = rti_info[RTAX_GATEWAY]) != NULL)
#elif __linux__
      if((sa = rti_info[RTF_GATEWAY]) != NULL)
#endif
	{
	  gw = (char *) sock_ntop_host(sa, sa->sa_family);
	
	  // if gw contains AF_LINK then dst can be reached in 1 hop
	  if (strncmp(gw, "AF_LINK", 7))
	    strncpy(next_hop, gw, 20);
	  else
	    strncpy(next_hop, dst, 20);
	}
      else
	{
	  WARNING ("Error 1");
	  status = ERROR;
	}
    }
  else
    {
      WARNING ("Error 2");
      status = ERROR;
    }

#ifdef DEBUG
  if ( (sa = rti_info[RTAX_DST]) != NULL)
    DEBUG ("dest: %s", sock_ntop_host(sa, sa->sa_family));
  if ( (sa = rti_info[RTAX_GATEWAY]) != NULL)
    DEBUG ("gateway: %s", sock_ntop_host(sa, sa->sa_family));
  if ( (sa = rti_info[RTAX_NETMASK]) != NULL)
    DEBUG ("netmask: %s", sock_masktop(sa, sa->sa_family));
  if ( (sa = rti_info[RTAX_GENMASK]) != NULL)
    DEBUF ("genmask: %s", sock_masktop(sa, sa->sa_family));
#endif

  close(sockfd);

  return status;
}
*/

#ifdef __FreeBSD__
// get addresses from routing table
void
get_routing_table_addr (int addrs, struct sockaddr *sa,
			struct sockaddr **rti_info)
{
  int i;
  for (i = 0; i < RTAX_MAX; i++)
    {
      if (addrs & (1 << i))
	{
	  rti_info[i] = sa;
	  NEXT_SA (sa);
	}
      else
	rti_info[i] = NULL;
    }
}
#endif

/* old version

// transform a socket adress 'sa' with length 'salen'
// into a readable IP address (string);
// return the IP address as a string
char *
sock_ntop_host (const struct sockaddr *sa, socklen_t salen)
{
  static char str[128];		// Unix domain is largest

  switch (sa->sa_family)
    {
    case AF_INET:
      {
	struct sockaddr_in *sin = (struct sockaddr_in *) sa;

	if (inet_ntop (AF_INET, &sin->sin_addr, str, sizeof (str)) == NULL)
	  return (NULL);
	return (str);
      }

    case AF_INET6:
      {
	struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *) sa;

	if (inet_ntop (AF_INET6, &sin6->sin6_addr, str, sizeof (str)) == NULL)
	  return (NULL);
	return (str);
      }

    case AF_LINK:
      {
	struct sockaddr_dl *sdl = (struct sockaddr_dl *) sa;

	if (sdl->sdl_nlen > 0)
	  snprintf (str, sizeof (str), "%*s",
		    sdl->sdl_nlen, &sdl->sdl_data[0]);
	else
	  snprintf (str, sizeof (str), "AF_LINK, index=%d", sdl->sdl_index);
	return (str);
      }

    default:
      snprintf (str, sizeof (str),
		"sock_ntop_host: unknown AF_xxx: %d, len %d", sa->sa_family,
		salen);
      return (str);
    }
  return (NULL);
}
*/

// transform a socket adress 'sa' with length 'salen'
// into a readable IP address (string);
// return the IP address as a string
char *sock_ntop_host(const struct sockaddr *sa, socklen_t salen)
{
  static char str[128]; // Unix domain is largest

  switch (sa->sa_family)
    {
    case AF_INET:
      {
	struct sockaddr_in*sin = (struct sockaddr_in *) sa;
	
	if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
	  return(NULL);
	return(str);
      }

    case AF_INET6:
      {
	struct sockaddr_in6*sin6 = (struct sockaddr_in6 *) sa;

	if (inet_ntop(AF_INET6, &sin6->sin6_addr, str, sizeof(str)) == NULL)
	  return(NULL);
	return(str);
      }

#if 0
  case AF_UNIX: {
    struct sockaddr_un*unp = (struct sockaddr_un *) sa;

    /* OK to have no pathname bound to the socket: happens on
       every connect() unless client calls bind() first. */
    if (unp->sun_path[0] == 0)
      strcpy(str, "(no pathname bound)");
    else
      snprintf(str, sizeof(str), "%s", unp->sun_path);
    return(str);
  }
#endif

#ifdef __FreeBSD__
    //case AF_LINK:
    case AF_PACKET:
      {
	struct sockaddr_dl*sdl = (struct sockaddr_dl *) sa;
	//struct sockaddr_dl*sll = (struct sockaddr_dl *) sa;
	
	if (sdl->sdl_nlen > 0)
	//if (sll->sll_family > 0)
	  snprintf(str, sizeof(str), "%*s",
		   sdl->sdl_nlen, &sdl->sdl_data[0]);
	else
	  snprintf(str, sizeof(str), "AF_LINK, index=%d", sdl->sdl_index);
	return(str);
      }
#endif

  default:
    snprintf(str, sizeof(str), "sock_ntop_host: unknown AF_xxx: %d, len %d",
             sa->sa_family, salen);
    return(str);
    }
  return (NULL);
}


// map an IP address value given as in_addr_t to internal QOMET node id;
// return id on success, ERROR on error
int
get_id_from_ip_in_addr_t (struct in_addr *IP_addresses, in_addr_t t)
{
  int i;

  for (i = 0; i < MAX_NODES_W; i++)
    {
      //INFO("Address %s: check %d and %d", s, ip, ipaddr[i]);
      if (IP_addresses[i].s_addr == t)
	return i;
    }

  return ERROR;
}

// map an IP address string to internal QOMET node id;
// return id on success, ERROR on error
int
get_id_from_ip_string (in_addr_t * IP_addresses, char *s)
{
  uint32_t ip;
  int i;

  if ((ip = inet_addr (s)) != INADDR_NONE)
    for (i = 0; i < MAX_NODES_W; i++)
      {
	//INFO("Address %s: check %d and %d", s, ip, ipaddr[i]);
	if (ip == IP_addresses[i])
	  return i;
      }
  else
    WARNING ("The address provided (%s) cannot be maped to an id", s);

  return ERROR;
}

// map an internal QOMET node id to an IP address; 
// the address is returned in the argument 'ip';
// return SUCCESS on success, ERROR on error
int
get_ip_from_id (char *IP_char_addresses, int id, char *ip)
{
  DEBUG ("Getting IP address for id '%d'...", id);

  if (id < 0 || id >= MAX_NODES_W)
    {
      WARNING ("Id %d is not within the valid range [%d,%d]",
	       id, 0, MAX_NODES_W);
      return ERROR;
    }

  strncpy (ip, IP_char_addresses + (id - FIRST_NODE_ID) * IP_ADDR_SIZE,
	   IP_ADDR_SIZE);

  DEBUG ("IP address for id '%d' is '%s'.", id, ip);

  return SUCCESS;
}

#ifdef __linux__

// Based on
//http://www.linuxjournal.com/article/8498?page=0,2

/*
void form_request_set(int rtl,  struct rtattr *rtap)
{
  // attributes of the route entry
  char dsts[24] = "192.168.0.100";
  int ifcn = 2, pn = 32;

  // initialize RTNETLINK request buffer
  bzero(&req, sizeof(req));

  // compute the initial length of the
  // service request
  rtl = sizeof(struct rtmsg);

  // add first attrib:
  // set destination IP addr and increment the
  // RTNETLINK buffer size
  rtap = (struct rtattr *) req.buf;
  rtap->rta_type = RTA_DST;
  rtap->rta_len = sizeof(struct rtattr) + 4;
  inet_pton(AF_INET, dsts,
     ((char *)rtap) + sizeof(struct rtattr));
  rtl += rtap->rta_len;

  // add second attrib:
  // set ifc index and increment the size
  rtap = (struct rtattr *) (((char *)rtap)
            + rtap->rta_len);
  rtap->rta_type = RTA_OIF;
  rtap->rta_len = sizeof(struct rtattr) + 4;
  memcpy(((char *)rtap) + sizeof(struct rtattr),
           &ifcn, 4);
  rtl += rtap->rta_len;

  // setup the NETLINK header
  req.nl.nlmsg_len = NLMSG_LENGTH(rtl);
  req.nl.nlmsg_flags = NLM_F_REQUEST | NLM_F_CREATE;
  req.nl.nlmsg_type = RTM_NEWROUTE;

  // setup the service header (struct rtmsg)
  req.rt.rtm_family = AF_INET;
  req.rt.rtm_table = RT_TABLE_MAIN;
  req.rt.rtm_protocol = RTPROT_STATIC;
  req.rt.rtm_scope = RT_SCOPE_UNIVERSE;
  req.rt.rtm_type = RTN_UNICAST;
  // set the network prefix size
  req.rt.rtm_dst_len = pn;
}
*/

void form_rtnetlink_request_get(struct rtnetlink_request *req)
{
  // initialize the request buffer
  bzero(req, sizeof(struct rtnetlink_request));

  // set the NETLINK header
  req->nl.nlmsg_len
             = NLMSG_LENGTH(sizeof(struct rtmsg));
  req->nl.nlmsg_flags = NLM_F_REQUEST | NLM_F_DUMP;
  req->nl.nlmsg_type = RTM_GETROUTE;

  // set the routing message header
  req->rt.rtm_family = AF_INET;
  req->rt.rtm_table = RT_TABLE_MAIN;
}


void send_rtnetlink_request(int fd, struct rtnetlink_request *req)
{
  struct sockaddr_nl pa;
  struct msghdr msg; 
  struct iovec iov; 
  int rtn;

  // create the remote address
  // to communicate
  bzero(&pa, sizeof(pa));
  pa.nl_family = AF_NETLINK;

  // initialize & create the struct msghdr supplied
  // to the sendmsg() function
  bzero(&msg, sizeof(msg));
  msg.msg_name = (void *) &pa;
  msg.msg_namelen = sizeof(pa);

  // place the pointer & size of the RTNETLINK
  // message in the struct msghdr
  bzero(&iov, sizeof(iov));
  iov.iov_base = (void *) &(req->nl);
  iov.iov_len = req->nl.nlmsg_len;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;

  // send the RTNETLINK message to kernel
  rtn = sendmsg(fd, &msg, 0);
  DEBUG ("Sent %d bytes to kernel.", rtn);
}


void recv_rtnetlink_reply(int fd, char *buf, int buf_size, int *nll)
{
  struct nlmsghdr *nlp;
  //struct sockaddr_nl la; 

  int rtn;

  char *p;

  // initialize the socket read buffer
  bzero(buf, buf_size);

  p = buf;
  (*nll) = 0;

  // read from the socket until the NLMSG_DONE is
  // returned in the type of the RTNETLINK message
  // or if it was a monitoring socket
  while(1)
    {
      rtn = recv(fd, p, buf_size - (*nll), 0);
      DEBUG ("Received %d bytes.", rtn);

      if (rtn == 0)
	break;

      nlp = (struct nlmsghdr *) p;

      if(nlp->nlmsg_type == NLMSG_DONE)
	break;

      // increment the buffer pointer to place
      // next message
      p += rtn;
      
      // increment the total size by the size of
      // the last received message
      (*nll) += rtn;

      /*      
      if((la.nl_groups & RTMGRP_IPV4_ROUTE)
	 == RTMGRP_IPV4_ROUTE)
	break;
      */
    }
}
/*
void read_reply(char *buf, int nll)
{
  struct rtmsg *rtp;
  struct rtattr *rtap;
  struct nlmsghdr *nlp;
  int rtl;

  // string to hold content of the route
  // table (i.e. one entry)
  char dsts[24], gws[24], ifs[16], ms[24];

  // outer loop: loops thru all the NETLINK
  // headers that also include the route entry
  // header
  nlp = (struct nlmsghdr *) buf;
  for(;NLMSG_OK(nlp, nll);nlp=NLMSG_NEXT(nlp, nll))
  {

    // get route entry header
    rtp = (struct rtmsg *) NLMSG_DATA(nlp);

    // we are only concerned about the
    // main route table
    if(rtp->rtm_table != RT_TABLE_MAIN)
      continue;

    // init all the strings
    bzero(dsts, sizeof(dsts));
    bzero(gws, sizeof(gws));
    bzero(ifs, sizeof(ifs));
    bzero(ms, sizeof(ms));

    // inner loop: loop thru all the attributes of
    // one route entry
    rtap = (struct rtattr *) RTM_RTA(rtp);
    rtl = RTM_PAYLOAD(nlp);
    for(;RTA_OK(rtap, rtl);rtap=RTA_NEXT(rtap,rtl))
    {
      switch(rtap->rta_type)
      {
        // destination IPv4 address
        case RTA_DST:
          inet_ntop(AF_INET, RTA_DATA(rtap),
                                     dsts, 24);
          break;

        // next hop IPv4 address
        case RTA_GATEWAY:
          inet_ntop(AF_INET, RTA_DATA(rtap),
                                     gws, 24);
          break;

        // unique ID associated with the network
        // interface
        case RTA_OIF:
          sprintf(ifs, "%d",
                   *((int *) RTA_DATA(rtap)));
        default:
          break;
      }
    }
    sprintf(ms, "%d", rtp->rtm_dst_len);

    if (strlen(dsts) == 0)
      strcpy(dsts, "0.0.0.0");

    if (strlen(gws) == 0)
      strcpy(gws, "0.0.0.0");

    DEBUG ("dst %s/%s gw %s if %s", dsts, ms, gws, ifs);
  }
}
*/

int
get_next_hop_ip_linux(char *dst_ip, char *next_hop_ip)
{
  struct rtmsg *rtp;
  struct rtattr *rtap;
  struct nlmsghdr *nlp;
  int rtl;

  // string to hold content of the route
  // table (i.e. one entry)
  char dsts[24], gws[24], ifs[16], ms[24];

  // variables used for
  // socket communications
  struct rtnetlink_request req;
  struct sockaddr_nl la;
  int fd;
  
  // buffer to hold the RTNETLINK reply(ies)
  char buf[BUFFER_SIZE];
  int nll;

  int status = SUCCESS;
  
  bzero(next_hop_ip, IP_ADDR_SIZE);

  // open socket
  DEBUG ("Opening socket...");
  fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
  if (fd == -1)
    {
      WARNING ("Could not open socket.");
      return ERROR;
    }

  // setup local address & bind using
  // this address
  bzero(&la, sizeof(la));
  la.nl_family = AF_NETLINK;
  la.nl_pid = getpid();

  // to monitor changes use this line!!!!!!!!
  //la.nl_groups = RTMGRP_IPV4_ROUTE | RTMGRP_NOTIFY;
  
  DEBUG ("Binding to socket...");
  if (bind(fd, (struct sockaddr*) &la, sizeof(la)) == -1)
    {
      WARNING ("Could not bind to socket.");
      status = ERROR;
    }

  // sub functions to create RTNETLINK message,
  // send over socket, receive reply & process
  // message
  DEBUG ("Forming request to get routing info...");
  form_rtnetlink_request_get(&req);

  DEBUG ("Sending routing info request...");
  send_rtnetlink_request(fd, &req);

  DEBUG ("Receiving routing info reply...");
  recv_rtnetlink_reply(fd, buf, BUFFER_SIZE, &nll);




  // start processing reply

  // outer loop: loops thru all the NETLINK
  // headers that also include the route entry
  // header
  nlp = (struct nlmsghdr *) buf;
  for(;NLMSG_OK(nlp, nll);nlp=NLMSG_NEXT(nlp, nll))
  {

    // get route entry header
    rtp = (struct rtmsg *) NLMSG_DATA(nlp);

    // we are only concerned about the
    // main route table
    if(rtp->rtm_table != RT_TABLE_MAIN)
      continue;

    // init all the strings
    bzero(dsts, sizeof(dsts));
    bzero(gws, sizeof(gws));
    bzero(ifs, sizeof(ifs));
    bzero(ms, sizeof(ms));

    // inner loop: loop thru all the attributes of
    // one route entry
    rtap = (struct rtattr *) RTM_RTA(rtp);
    rtl = RTM_PAYLOAD(nlp);
    for(;RTA_OK(rtap, rtl);rtap=RTA_NEXT(rtap,rtl))
    {
      switch(rtap->rta_type)
      {
        // destination IPv4 address
        case RTA_DST:
          inet_ntop(AF_INET, RTA_DATA(rtap),
                                     dsts, 24);
          break;

        // next hop IPv4 address
        case RTA_GATEWAY:
          inet_ntop(AF_INET, RTA_DATA(rtap),
                                     gws, 24);
          break;

        // unique ID associated with the network
        // interface
        case RTA_OIF:
          sprintf(ifs, "%d",
                   *((int *) RTA_DATA(rtap)));
        default:
          break;
      }
    }
    sprintf(ms, "%d", rtp->rtm_dst_len);

    // identify the right entry
    if (strncmp(dst_ip, dsts, IP_ADDR_SIZE) == 0)
      {
	// if gateway was identified, copy it
	if (strlen(gws) != 0)
	  strncpy(next_hop_ip, gws, IP_ADDR_SIZE);
	// otherwise use the destination itself as gateway (next hop)
	else
	  strncpy(next_hop_ip, dst_ip, IP_ADDR_SIZE);
      }

    // beautify for printing
    if (strlen(dsts) == 0)
      strcpy(dsts, "0.0.0.0");
    if (strlen(gws) == 0)
      strcpy(gws, "0.0.0.0");

    DEBUG ("dst %s/%s gw %s if %s", dsts, ms, gws, ifs);
  }

  // if no gateway was identified, copy the destination itself as
  // next hop
  if (strlen(next_hop_ip) == 0)
    strncpy(next_hop_ip, dst_ip, IP_ADDR_SIZE);

  DEBUG ("Next hop for destination '%s' is '%s'", dst_ip, next_hop_ip);
  // end processing reply




  // close socket
  DEBUG ("Closing socket...");
  if (close(fd) == -1)
    {
      WARNING ("Could not close socket.");
      status = ERROR;
    }

  return status;
}

#endif
