
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 *
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: test_routing_info.c
 * Function: Source file for testing routing_info.c.
 *
 * Authors: Razvan Beuran
 *
 *   $Revision: 128 $
 *   $LastChangedDate: 2009-02-06 10:21:50 +0900 (Fri, 06 Feb 2009) $
 *   $LastChangedBy: razvan $
 *
 ***********************************************************************/

#include "wireconf.h"

////////////////////////////////////////////////////
// Test parameters
////////////////////////////////////////////////////

// NOTE: the IP addresses in the settings file must be valid for your system. 
// In particular, the gateway (next hop) must be reachable in order
// for the route commands to work. Moreover, the ids given below must be 
// present in the settings file.

char SETTINGS_FILE[]="test_routing_info.settings";

int ID_SELF = 0;

// test 1: next hop is the same with the destination
int ID_DEST1 = 1;
int ID_NEXT1 = 1;

// test 2: next hop is the different from the destination
int ID_DEST2 = 2;
int ID_NEXT2 = 1;


////////////////////////////////////////////////////
// Other constants
////////////////////////////////////////////////////

unsigned SLEEP_TIME = 1;


////////////////////////////////////////////////////
// Helper functions
////////////////////////////////////////////////////

// add a route
int
add_route(int id_dest, char *ip_dest, char *ip_next)
{
  char cmd[MAX_STRING];
  int status;

  INFO ("Adding route information for destination '%d' (IP='%s'): \
next hop has IP='%s'...", id_dest, ip_dest, ip_next);
  snprintf (cmd, MAX_STRING-1, "route add %s gw %s", ip_dest, ip_next);
  DEBUG ("Making system call: '%s'", cmd);
  status = system (cmd);

  if (status != 0)
    {
      WARNING("Could not add route. Make sure the test settings are correct \
(e.g., the next hop is reachable). NOTE: The test requires root privileges.");
      return ERROR;
    }

  return SUCCESS;
}

// delete a route
int
delete_route(int id_dest, char *ip_dest)
{
  char cmd[MAX_STRING];
  int status;

  INFO ("Deleting route information for destination '%d' (IP='%s')...", 
	id_dest, ip_dest);
  snprintf (cmd, MAX_STRING-1, "route delete %s", ip_dest);
  DEBUG ("Making system call: '%s'", cmd);
  status = system (cmd);

  if (status != 0)
    {
      WARNING("Could not delete route. NOTE: The test requires root \
privileges.");
      return ERROR;
    }

  return SUCCESS;
}

int
do_get_next_hop_id(in_addr_t * IP_addresses, char *IP_char_addresses,
		   int dst_id, int direction, int expected_next_hop_id)
{
  // return status
  int status = SUCCESS;

  // id of next hop
  int next_hop_id;

  // get next hop id
  INFO ("Getting next hop id for destination '%d' (direction=%s)...", dst_id,
	(direction==DIRECTION_OUT)?"OUT":
	((direction==DIRECTION_IN)?"IN":"BOTH"));
  next_hop_id = 
    get_next_hop_id(IP_addresses, IP_char_addresses, dst_id, direction);
  if (next_hop_id != ERROR)
    {
      INFO ("\tNext hop for destination '%d' has id '%d': CORRECT", 
	    dst_id, next_hop_id);
      if (next_hop_id != expected_next_hop_id)
	{
	  WARNING ("Incorrect next hop id for destination '%d' (returned \
result is '%d', expected result was '%d').", ID_DEST1, 
		   next_hop_id, expected_next_hop_id);
	  status = ERROR;
	}
    }
  else
    {
      WARNING ("Failed to get next hop id for destination '%d'...", dst_id);
      status = ERROR;
    }

  return status;
}


////////////////////////////////////////////////////
// Main function of the program
////////////////////////////////////////////////////

int
main()
{
  // array of IP adresses of all nodes in the experiments
  in_addr_t IP_addresses[MAX_NODES_W];

  // array of IP adresses of all nodes in the experiments
  // (string form)
  char IP_char_addresses[MAX_NODES_W * IP_ADDR_SIZE];

  char *IP_DEST1, *IP_NEXT1;
  char *IP_DEST2, *IP_NEXT2;

  // number of nodes
  int node_count;

  // node iterator
  //int node_i;

  // program return status
  int status;

  

  ////////////////////////////////////////////////////
  // Test header

  // print test settings
  INFO ("Testing routing_info implementation in the wireconf library...");
  INFO ("Using settings in file '%s'.", SETTINGS_FILE);

  // read IP settings
  if ((node_count = io_read_settings_file (SETTINGS_FILE, 
					   IP_addresses, IP_char_addresses,
					   MAX_NODES_W)) < 1)
    {
      WARNING ("Invalid setting file: '%s'", SETTINGS_FILE);
      return ERROR;
    }
  INFO ("\tRead IP settings for %d nodes.", node_count);

  // initialize IP addresses for each test
  IP_DEST1 = IP_char_addresses + ID_DEST1 * IP_ADDR_SIZE;
  IP_NEXT1 = IP_char_addresses + ID_NEXT1 * IP_ADDR_SIZE;
  IP_DEST2 = IP_char_addresses + ID_DEST2 * IP_ADDR_SIZE;
  IP_NEXT2 = IP_char_addresses + ID_NEXT2 * IP_ADDR_SIZE;

  
  ////////////////////////////////////////////////////
  // Test initialization

  // adding rules for test 1
  if (add_route(ID_DEST1, IP_DEST1, IP_NEXT1) == ERROR)
    status = ERROR;

  // adding rules for test 2
  if (add_route(ID_DEST2, IP_DEST2, IP_NEXT2) == ERROR)
    status = ERROR;


  ////////////////////////////////////////////////////
  // Just sleeping for a while

  INFO ("Sleeping %u seconds...", SLEEP_TIME);
  sleep(SLEEP_TIME);


  ////////////////////////////////////////////////////
  // Testing retrieval of next hop id

  // get next hop for test 0
  if (do_get_next_hop_id(IP_addresses, IP_char_addresses, 
			 ID_SELF, DIRECTION_OUT, ID_SELF) == ERROR)
    status = ERROR;

  // get next hop for test 1
  if (do_get_next_hop_id(IP_addresses, IP_char_addresses, 
			 ID_DEST1, DIRECTION_OUT, ID_NEXT1) == ERROR)
    status = ERROR;

  // get next hop for test 2a
  if (do_get_next_hop_id(IP_addresses, IP_char_addresses, 
			 ID_DEST2, DIRECTION_OUT, ID_NEXT2) == ERROR)
    status = ERROR;

  /* is DIRECTION_IN next hop meaningful?!; equivalent to previous hop?!
  // get next hop for test 2b
  if (do_get_next_hop_id(IP_addresses, IP_char_addresses, 
			 ID_DEST2, DIRECTION_IN, ID_DEST2) == ERROR)
    status = ERROR;
  */


  ////////////////////////////////////////////////////
  // Test finalization

  // deleting rules for test 1
  if (delete_route(ID_DEST1, IP_DEST1) == ERROR)
    status = ERROR;

  // deleting rules for test 2
  if (delete_route(ID_DEST2, IP_DEST2) == ERROR)
    status = ERROR;

  // check whether any errors occurred
  if (status == ERROR)
    WARNING ("Errors encountered during execution. Make sure the test \
settings are correct.");

  return SUCCESS;
}
