
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 * 
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: wireconf.c
 * Function: Main source file of the real-time QOMET component 
 *           that effectively performs the network emulation. 
 *
 * Authors: Junya Nakata, Lan Nguyen Tien, Razvan Beuran
 *
 *   $Revision: 128 $
 *   $LastChangedDate: 2009-02-06 10:21:50 +0900 (Fri, 06 Feb 2009) $
 *   $LastChangedBy: razvan $
 *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>

#include "../deltaQ/deltaQ.h"
#include "../timer/timer.h"
#include "wireconf.h"


///////////////////////////////////////////////////////
// Flag to enable deltaQ adjustment based on contention
// (turned ON by default, turn OFF only for debugging)
///////////////////////////////////////////////////////

#define ADJUST_DELTAQ


/////////////////////////////////////////////
// Basic constants
/////////////////////////////////////////////

// controlling pipe ids
int MIN_PIPE_ID_IN        = 10000;
int MIN_PIPE_ID_OUT       = 20000;
int MIN_PIPE_ID_IN_BCAST  = 30000;
int MIN_PIPE_ID_OUT_BCAST = 40000;

// maximum number of rules
int MAX_RULE_COUNT	  = 500;

// default frame size
int DEFAULT_FRAME_SIZE    = 1024;

// used to signify "undefined" for numbers
int UNDEFINED_SIGNED      = -1;
unsigned int UNDEFINED_UNSIGNED    = 65535;
float UNDEFINED_BANDWIDTH = -1.0; 

/////////////////////////////////////////////
// Scaling factor
/////////////////////////////////////////////

// scaling factor to slow down (or speed up) execution if desired;
// the value 1.0 means "no scaling" (normal execution); 
// values larger than 1.0 mean "slow down" execution;
// values smaller than 1.0 mean "speed up" execution
#ifndef SCALING_FACTOR
#define SCALING_FACTOR          1.0	
#endif


///////////////////////////////////
// Generic functions
///////////////////////////////////

// print a brief usage of this program
void
usage (char *argv0)
{
  INFO ("\ndo_wireconf. Drive network emulation using QOMET data.\n");
  INFO ("Usage: %s -q <qomet_output.bin>\n"
	   "\t\t     -i <current_id>\n"
	   "\t\t     -s <settings_file>\n"
	   "\t\t     [-b <broadcast_address>]\n", argv0);
  INFO ("do_wireconf uses the binary QOMET data in <qomet_output.bin> \
to configure");
  INFO ("all connections from the current node <current_id> given the \
IP settings");
  INFO ("in <settings_file> and the OPTIONAL broadcast address \
<broadcast_address>.");
  INFO ("Each line of the settings file contains pairs of node ids \
and IP addresses.\n");
}

void
delete_all_rules_and_pipes(int my_id, int dummynet_socket_id, int node_count,
			   int do_trigger_error)
{
  int j;
    
  // delete the unicast dummynet rules
  for (j = FIRST_NODE_ID; j < node_count + FIRST_NODE_ID; j++)
    {
      if (j == my_id)
	continue;

      DEBUG ("Deleting rule #%d...", MIN_PIPE_ID_OUT + j);
      if (delete_rule (dummynet_socket_id, MIN_PIPE_ID_OUT + j) == ERROR)
	{
	  WARNING ("Could not delete rule #%d", MIN_PIPE_ID_OUT + j);
	}
      DEBUG ("Deleting pipe #%d...", MIN_PIPE_ID_OUT + j);
      if (delete_pipe (dummynet_socket_id, MIN_PIPE_ID_OUT + j) == ERROR)
	{
	  WARNING ("Could not delete pipe #%d", MIN_PIPE_ID_OUT + j);
	}
    }

  //delete broadcast dummynet rules
  for (j = FIRST_NODE_ID; j < node_count + FIRST_NODE_ID; j++)
    {
      if (j == my_id)
	continue;

      DEBUG ("Deleting rule #%d...", MIN_PIPE_ID_IN_BCAST + j);
      if (delete_rule (dummynet_socket_id, MIN_PIPE_ID_IN_BCAST + j)
	  == ERROR)
	{
	  WARNING ("Could not delete rule #%d", MIN_PIPE_ID_IN_BCAST + j);
	}
      DEBUG ("Deleting pipe #%d...", MIN_PIPE_ID_IN_BCAST + j);
      if (delete_pipe (dummynet_socket_id, MIN_PIPE_ID_IN_BCAST + j)
	  == ERROR)
	{
	  WARNING ("Could not delete pipe #%d", MIN_PIPE_ID_IN_BCAST + j);
	}
    }
}


///////////////////////////////////
// Main function of the program
///////////////////////////////////

int
main (int argc, char *argv[])
{

  /////////////////////////////////////////////////////
  // General variables

  // deltaQ paramters to be applied to dummynet
  float bandwidth, loss_rate, delay;

  // time for the current records
  float crt_record_time = 0.0;

  // qomet file descriptor
  FILE *qomet_fd = NULL;

  // timer handle for regulating Dummynet configuration
  struct timer_handle timer;

  // loop count used in predefined experiments
  int loop_count = 0;

  // time at the beginning and end of the experiment; 
  // used to measure total execution time
  struct timeval tp_begin, tp_end;

  // character used when parsing command line options
  char option_ch;

  // dummynet control socket id
  int dummynet_socket_id;

  // the id of this node
  int my_id = UNDEFINED_SIGNED;

  // array of next hop ids
  int *next_hop_ids = NULL;

  // array of IP adresses of all nodes in the experiments
  in_addr_t IP_addresses[MAX_NODES_W];

  // array of IP adresses of all nodes in the experiments
  // (string form)
  char IP_char_addresses[MAX_NODES_W * IP_ADDR_SIZE];

  // broadcast address as string
  char broadcast_address[IP_ADDR_SIZE];

  // number of nodes in the experiments
  // (determined from settings files)
  int node_count = UNDEFINED_SIGNED;

  // iterator for nodes
  int node_i;

  // memory to store rule data as a block
  struct ip_fw *rule_data = NULL;

  // size of allocate memory for rule data
  int rule_data_alloc_size = UNDEFINED_ALLOC_SIZE;

  // arrary of pointers to rule data
  struct ip_fw *rules[MAX_RULE_COUNT];

  // effective number of rules  
  int rule_count;

  // arrays of counters for the number of packets/bytes
  // transmitted to each destination during the last
  // time interval
  long long unsigned int *last_pkt_counters = NULL;
  long long unsigned int *last_byte_counters = NULL;

  // array of average frame size statistics
  float *avg_frame_sizes = NULL;

  // wireconf structure
  struct wireconf_class wireconf;

  // configuration file header data structure
  struct binary_header_class binary_header;

  // time records counter
  long int time_i;

  // dummynet configuration records
  struct binary_time_record_class binary_time_record;
  struct binary_record_class *binary_records = NULL;
  int binary_records_max_count;

  struct binary_record_class **my_records_ucast = NULL;
  struct binary_record_class *adjusted_records_ucast = NULL;
  int *my_records_ucast_changed = NULL;
  struct binary_record_class *my_records_bcast = NULL;
  int *my_records_bcast_changed = NULL;

  // channel utilizations
  float my_total_channel_utilization;
  float my_total_transmission_probability;

  int pkt_i;

  // flag indicating that the experiment is local, hence no routing
  // information is to be collected
  int local_experiment;
  

  /////////////////////////////////////////////////////
  // Basic variable initialization

  // default broadcast address
  strncpy (broadcast_address, "255.255.255.255", IP_ADDR_SIZE);


  /////////////////////////////////////////////////////
  // Argument parsing

  // basic check for arguments
  if (argc < 2)
    {
      WARNING ("No arguments provided");
      usage (argv[0]);
      exit (1);
    }

  // parse command-line options
  while ((option_ch = getopt (argc, argv, "q:i:s:b:h")) != -1)
    {
      //int node_i;

      switch (option_ch)
	{
	  //QOMET output file
	case 'q':
	  if ((qomet_fd = fopen (optarg, "r")) == NULL)
	    {
	      WARNING ("Could not open QOMET output file '%s'", optarg);
	      exit (1);
	    }
	  else
	    INFO ("Using QOMET file: %s", optarg);
	  break;

	  // current node ID
	case 'i':
	  my_id = strtol (optarg, NULL, 10);
	  break;

	  // settings file
	case 's':
	  if ((node_count
	       = io_read_settings_file (optarg, IP_addresses, IP_char_addresses,
					MAX_NODES_W)) < 1)
	    {
	      WARNING ("Invalid setting file: '%s'", optarg);
	      exit (1);
	    }
	  break;

	case 'b':
	  strncpy (broadcast_address, optarg, IP_ADDR_SIZE);
	  break;

	  // help output
	case 'h':
	case '?':
	default:
	  usage (argv[0]);
	  exit (1);
	}
    }

  if ((my_id < FIRST_NODE_ID) || (my_id >= node_count + FIRST_NODE_ID))
    {
      WARNING ("Invalid node id: '%d'. Valid range is [%d, %d]", my_id,
	       FIRST_NODE_ID, node_count + FIRST_NODE_ID - 1);
      exit (1);
    }

  // update argument-related counter and pointer
  argc -= optind;
  argv += optind;

  // check that all the required arguments were provided
  if (qomet_fd == NULL)
    {
      WARNING ("No QOMET data file was provided");
      usage (argv[0]);
      exit (1);
    }

  if ((my_id == UNDEFINED_SIGNED) || (node_count == UNDEFINED_SIGNED))
    {
      WARNING ("Insufficient arguments were provided");
      usage (argv[0]);
      fclose (qomet_fd);
      exit (1);
    }


  /////////////////////////////////////////////////////
  // More initialization

  // detect if the test is local by comparing IP of current node
  // with loopback IP
  if (strncmp(IP_char_addresses + (my_id - FIRST_NODE_ID) * IP_ADDR_SIZE, 
	      "127.0.0.", 8) == 0)
    {
      INFO ("Current node has loopback address => local experiment (deltaQ \
adjustment based on statistics not used).");
      local_experiment = TRUE;
    }
  else
    local_experiment = FALSE;

  // initialize wireconf library
  if (wireconf_init (&wireconf, my_id, IP_addresses, node_count) == ERROR)
    {
      WARNING ("Could not initialize wireconf library");
      exit (1);
    }

  // open dummynet socket
  DEBUG ("Open dummynet socket...");
  if ((dummynet_socket_id = get_socket ()) < 0)
    {
      WARNING ("Could not open dummynet socket (requires root priviledges)");
      exit (1);
    }


  /////////////////////////////////////////////////////
  // Prepare dummynet rules/pipes

  INFO ("Cleaning up older rules and pipes from dummynet (warnings can be \
ignored)...");
  delete_all_rules_and_pipes(my_id, dummynet_socket_id, node_count, FALSE);

  INFO ("Adding rules and pipes to ipfw/dummynet...");

  int j;

  // add rule & pipe for unicast traffic _to_ j
  for (j = FIRST_NODE_ID; j < node_count + FIRST_NODE_ID; j++)
    {
      if (j == my_id)
	continue;

      DEBUG ("Node %d: Adding rule #%d with pipe #%d to destination \
%s (OUT)...", my_id, MIN_PIPE_ID_OUT + j, MIN_PIPE_ID_OUT + j, 
	     IP_char_addresses + (j - FIRST_NODE_ID) * IP_ADDR_SIZE);
      
      if (add_rule_and_pipe
	  (dummynet_socket_id, MIN_PIPE_ID_OUT + j, MIN_PIPE_ID_OUT + j,
	   "any", IP_char_addresses + (j - FIRST_NODE_ID) * IP_ADDR_SIZE,
	   DIRECTION_OUT) < 0)
	{
	  WARNING ("Node %d: Could not add rule #%d with pipe #%d to \
destination %s (OUT)", my_id, MIN_PIPE_ID_OUT + j, MIN_PIPE_ID_OUT + j, 
		   IP_char_addresses + (j - FIRST_NODE_ID) * IP_ADDR_SIZE);
	  exit (1);
	}
    }			// end for loop
  
  // add rule & pipe for broadcast traffic _from_ j
  for (j = FIRST_NODE_ID; j < node_count + FIRST_NODE_ID; j++)
    {
      if (j == my_id)
	continue;
      
      DEBUG ("Node %d: Adding rule #%d with pipe #%d to destination \
%s (IN)...", my_id, MIN_PIPE_ID_IN_BCAST + j, MIN_PIPE_ID_IN_BCAST + j, 
	     broadcast_address);
      if (add_rule_and_pipe
	  (dummynet_socket_id, MIN_PIPE_ID_IN_BCAST + j,
	   MIN_PIPE_ID_IN_BCAST + j,
	   IP_char_addresses + (j - FIRST_NODE_ID) *
	   IP_ADDR_SIZE, broadcast_address, DIRECTION_IN) < 0)
	{
	  WARNING ("Node %d: Could not add rule #%d with pipe #%d from %s \
to destination %s (IN)", my_id, MIN_PIPE_ID_IN_BCAST + j,
		   MIN_PIPE_ID_IN_BCAST + j, IP_char_addresses 
		   + (j - FIRST_NODE_ID) * IP_ADDR_SIZE, 
		   broadcast_address);
	  exit (1);
	}
    }


  ///////////////////////////////////////////////////////////
  // Read QOMET binary file header and do memory allocations

  INFO ("Start reading binary QOMET output file...");

  // read and check binary header
  if (io_binary_read_header_from_file (&binary_header, qomet_fd) == ERROR)
    {
      WARNING ("Aborting on input error (binary header)");
      exit (1);
    }
  io_binary_print_header (&binary_header);

  if (node_count != binary_header.interface_number)
    {
      WARNING ("Number of nodes according to the settings file (%d) and \
number of nodes according to QOMET scenario (%d) differ", 
	       node_count, binary_header.interface_number);
      exit (1);
    }

  // compute the maximum number of records we can expect
  binary_records_max_count = binary_header.interface_number
    * (binary_header.interface_number - 1);

  // allocate memory for binary records
  binary_records = (struct binary_record_class *)
    calloc (binary_records_max_count, sizeof (struct binary_record_class));
  if (binary_records == NULL)
    {
      WARNING ("Cannot allocate memory for binary records");
      exit (1);
    }

  // allocate memory for next hop ids
  next_hop_ids = (int *) calloc (binary_header.interface_number, sizeof (int));
  if (next_hop_ids == NULL)
    {
      WARNING ("Cannot allocate memory for next_hop_ids");
      exit (1);
    }

  // allocate memory for my records when using UCAST
  my_records_ucast = (struct binary_record_class **)
    calloc (binary_header.interface_number, sizeof (struct binary_record_class *));
  if (my_records_ucast == NULL)
    {
      WARNING ("Cannot allocate memory for my_records_ucast");
      exit (1);
    }
  for (node_i = 0; node_i < binary_header.interface_number; node_i++)
    {
      int j;

      my_records_ucast[node_i] = (struct binary_record_class *)
	calloc (binary_header.interface_number, sizeof (struct binary_record_class));

      if (my_records_ucast[node_i] == NULL)
	{
	  WARNING ("Cannot allocate memory for my_records_ucast[%d]", node_i);
	  exit (1);
	}
      
      for (j = 0; j < binary_header.interface_number; j++)
	my_records_ucast[node_i][j].bandwidth = UNDEFINED_BANDWIDTH;
    }

  // allocate memory for adjusted records when using UCAST
  adjusted_records_ucast = (struct binary_record_class *)
    calloc (binary_header.interface_number, sizeof (struct binary_record_class));
  if (adjusted_records_ucast == NULL)
    {
      WARNING ("Cannot allocate memory for adjusted_records_ucast");
      exit (1);
    }

  // allocate memory for my records changed when using UCAST
  my_records_ucast_changed = (int *) calloc (binary_header.interface_number,
					     sizeof (int));
  if (my_records_ucast_changed == NULL)
    {
      WARNING ("Cannot allocate memory for my_records_ucast_changed");
      exit (1);
    }

  // allocate memory for my records when using BCAST
  my_records_bcast =
    (struct binary_record_class *) calloc (binary_header.interface_number,
				    sizeof (struct binary_record_class));
  if (my_records_bcast == NULL)
    {
      WARNING ("Cannot allocate memory for my_records_bcast");
      exit (1);
    }
  for (node_i = 0; node_i < binary_header.interface_number; node_i++)
    my_records_bcast[node_i].bandwidth = UNDEFINED_BANDWIDTH;

  // allocate memory for my records changed when using BCAST
  my_records_bcast_changed = (int *) calloc (binary_header.interface_number,
					     sizeof (int));
  if (my_records_bcast_changed == NULL)
    {
      WARNING ("Cannot allocate memory for my_records_bcast_changed");
      exit (1);
    }

  // allocate memory for last byte counters
  last_byte_counters = (long long unsigned int *)
    calloc (binary_header.interface_number, sizeof (long long unsigned int));
  if (last_byte_counters == NULL)
    {
      WARNING ("Cannot allocate memory for last_byte_counters");
      exit (1);
    }

  // allocate memory for last pkt counters
  last_pkt_counters = (long long unsigned int *)
    calloc (binary_header.interface_number, sizeof (long long unsigned int));
  if (last_pkt_counters == NULL)
    {
      WARNING ("Cannot allocate memory for last_pkt_counters");
      exit (1);
    }

  // allocate memory for avg frame size array
  avg_frame_sizes = (float *)
    calloc (binary_header.interface_number, sizeof (float));
  if (avg_frame_sizes == NULL)
    {
      WARNING ("Cannot allocate memory for avg_frame_sizes");
      exit (1);
    }

  for (pkt_i = 0; pkt_i < binary_header.interface_number; pkt_i++)
    {
      //last_pkt_counters[pkt_i] = 0; // calloc inits to 0 => not needed!
      //last_byte_counters[pkt_i] = 0;
      avg_frame_sizes[pkt_i] = DEFAULT_FRAME_SIZE;
    }


  //////////////////////////////////////////////////
  // Start experiment

  // get the time at the beginning of the experiment
  gettimeofday (&tp_begin, NULL);

  // loop through all time records
  for (time_i = 0; time_i < binary_header.time_record_number; time_i++)
    {
      // records counter
      int rec_i;

      // read input data from QOMET output file
      DEBUG ("Reading QOMET data from file...");

      // read time record
      if (io_binary_read_time_record_from_file (&binary_time_record,
						qomet_fd) == ERROR)
	{
	  WARNING ("Aborting on input error (time record)");
	  exit (1);
	}
      io_binary_print_time_record (&binary_time_record);

      // get time from record
      crt_record_time = binary_time_record.time;

      // check number of records to be read doesn't exceed the maximum
      if (binary_time_record.record_number > binary_records_max_count)
	{
	  WARNING
	    ("The number of records to be read exceeds allocated size (%d)",
	     binary_records_max_count);
	  exit (1);
	}

      // read binary records in an array
      if (io_binary_read_records_from_file (binary_records,
					    binary_time_record.record_number,
					    qomet_fd) == ERROR)
	{
	  WARNING ("Aborting on input error (records)");
	  exit (1);
	}

      // loop through all binary records
      for (rec_i = 0; rec_i < binary_time_record.record_number; rec_i++)
	{
#ifdef DEBUG
	  //io_binary_print_record (&(binary_records[rec_i]));
#endif

	  // check source is in valid range
	  if (binary_records[rec_i].from_id < FIRST_NODE_ID
	      || (binary_records[rec_i].from_id
		  > binary_header.interface_number + FIRST_NODE_ID - 1))
	    {
	      INFO ("Source with id = %d is out of the valid \
range [%d, %d]", binary_records[rec_i].from_id, FIRST_NODE_ID, 
		    binary_header.interface_number + FIRST_NODE_ID - 1);
	      exit (1);
	    }

	  // check destination is in valid range
	  if (binary_records[rec_i].to_id < FIRST_NODE_ID
	      || (binary_records[rec_i].to_id
		  > binary_header.interface_number + FIRST_NODE_ID - 1))
	    {
	      INFO ("Destination with id = %d is out of the valid \
range [%d, %d]", binary_records[rec_i].to_id, FIRST_NODE_ID, 
		    binary_header.interface_number + FIRST_NODE_ID - 1);
	      exit (1);
	    }

	  // copy records to the appropriate storage slot,
	  // and set the changed flags
	  if (binary_records[rec_i].from_id == my_id)
	    {
	      io_binary_copy_record
		(&(my_records_ucast[binary_records[rec_i].from_id]
		   [binary_records[rec_i].to_id]), &binary_records[rec_i]);
	      my_records_ucast_changed[binary_records[rec_i].to_id] = TRUE;
#ifdef DEBUG
	      DEBUG ("Copied binary_records to my_records_ucast (index is binary_records[rec_i].to_id=%d).", binary_records[rec_i].to_id);
	      //io_binary_print_record (&(my_records_ucast[binary_records[rec_i].from_node][binary_records[rec_i].to_node]));
#endif
	    }

	  // check whether the to_node matches my id
	  // (data will be used in BCAST configuration)
	  if (binary_records[rec_i].to_id == my_id)
	    {
	      // copy records to the appropriate storage slot,
	      // and set the changed flags
	      io_binary_copy_record
		(&(my_records_bcast[binary_records[rec_i].from_id]),
		 &binary_records[rec_i]);
	      my_records_bcast_changed[binary_records[rec_i].from_id] =
		TRUE;
#ifdef DEBUG
	      DEBUG ("Copied binary_records to my_records_bcast (index is \
binary_records[rec_i].from_node=%d).", binary_records[rec_i].from_id);
	      //io_binary_print_record (&(my_records_bcast[binary_records[rec_i].from_node]));
#endif
	    }
	}

      // when there are no records, conditions could still change if the
      // channel utilization for other nodes changed => could check this
      // and only go further if any change took place
      if (binary_time_record.record_number == 0)
	{
	  // NOT IMPLEMENTED YET
	}

      // if it is the first iteration, copy values to adjusted array
      if (time_i == 0)
	{
	  // copy deltaQ
	  for (rec_i = 0; rec_i < node_count; rec_i++)
	    // do not consider the node itself
	    if (rec_i != my_id)
	      {
		//copy first all fields of the record
		io_binary_copy_record (&(adjusted_records_ucast[rec_i]),
				       &(my_records_ucast[my_id][rec_i]));
#ifdef DEBUG
		DEBUG ("Copied my_records_ucast to adjusted_records_ucast \
(index is rec_i=%d).", rec_i);
		//io_binary_print_record (&(adjusted_records_ucast[rec_i]));
#endif
	      }
	}
      // if it is not the first iteration
      else
	{
	  // adjust deltaQ parameters given the real traffic conditions
	  // of the _other_nodes_ as computed at the previous step
	  // (contention from others), as well as the self contention
	  // caused by the node itself;
	  // adjusted_records_ucast will contain the output data
	  
	  // TEMPORARY!!!!!!!!!!!!!!!!!!!!!!!!!

	  // use one of the blocks below depending on whether deltaQ adjustment
	  // should be disabled or enabled

#ifndef ADJUST_DELTAQ
	  WARNING ("Adjustment of deltaQ is disabled.");

	  // copy deltaQ
	  for (rec_i = 0; rec_i < node_count; rec_i++)
	    // do not consider the node itself
	    if (rec_i != my_id)
	      {
		//copy first all fields of the record
		io_binary_copy_record (&(adjusted_records_ucast[rec_i]),
				       &(my_records_ucast[my_id][rec_i]));
#ifdef DEBUG
		DEBUG ("Copied my_records_ucast to adjusted_records_ucast \
(index is rec_i=%d).", rec_i);
		//io_binary_print_record (&(adjusted_records_ucast[rec_i]));
#endif
	      }
#else
	  DEBUG ("Adjustment of deltaQ is enabled.");
	  adjust_deltaQ (&wireconf, my_records_ucast,
			 adjusted_records_ucast, my_records_ucast_changed,
			 avg_frame_sizes);
#endif
	}


      // waiting for the next timer deadline event;
      // if this is the first operation we reset the timer,
      // and proceed without waiting
      if (time_i == 0)
	timer_reset (&timer, crt_record_time);
      else
	{
	  if (SCALING_FACTOR == 1.0)
	    INFO ("Waiting to reach time %.2f s...", crt_record_time);
	  else
	    INFO ("Waiting to reach real time %.2f s (scenario \
time %.2f)...", crt_record_time * SCALING_FACTOR, crt_record_time);

	  // wait for for the next timer event
	  timer_wait (&timer, crt_record_time * SCALING_FACTOR);
	}

      // when timer event comes, apply configuration for all 
      // the unicast links with destination 'node_i'
      // NOTE: GETTING NEXT HOP COULD BE DONE WHILE WAITING
      for (node_i = FIRST_NODE_ID; node_i < (node_count + FIRST_NODE_ID);
	   node_i++)
	{
	  // id of the next hop
	  int next_hop_id;

	  // skip current node
	  if (node_i == my_id)
	    continue;

	  // get the next hop ID for destination node_i
	  next_hop_id = get_next_hop_id (IP_addresses, IP_char_addresses, 
					 node_i, DIRECTION_OUT);
	  if (next_hop_id == ERROR)
	    {
	      WARNING ("Could not locate the next hop for destination node %i",
		       node_i);
	      next_hop_id = node_i;
	    }
	  DEBUG ("Next_hop=%i for destination=%i", next_hop_id, node_i);

	  if (next_hop_id < 0
	      || (next_hop_id > binary_header.interface_number - 1))
	    {
	      WARNING ("Next hop with id = %d is out of the valid \
range [%d, %d]", binary_records[rec_i].to_id, 0, 
		       binary_header.interface_number - 1);
	      exit (1);
	    }

	  // save next_hop_id
	  next_hop_ids[node_i] = next_hop_id;

	  /* IGNORE FLAG FOR THE MOMENT

	  // if the parameters for the next hop did not change, 
	  // there is no need to proceed any further
	  if (my_records_ucast_changed[next_hop_id] == FALSE)
	  continue;
	  // otherwise, set the flag to FALSE and proceed 
	  else
	  my_records_ucast_changed[next_hop_id] = FALSE;
	  */
	  bandwidth = adjusted_records_ucast[next_hop_id].bandwidth;
	  delay = adjusted_records_ucast[next_hop_id].delay;
	  loss_rate = adjusted_records_ucast[next_hop_id].loss_rate;
#ifdef DEBUG
	  DEBUG ("Used contents of adjusted_records_ucast for deltaQ \
parameters (index is next_hop_id=%d).", next_hop_id);
	  //io_binary_print_record (&(adjusted_records_ucast[next_hop_id]));
#endif

	  if (bandwidth != UNDEFINED_BANDWIDTH)
	    {
	      // print current configuration info
	      INFO ("-- Wireconf pipe=%d: #%d UCAST to #%d (next_hop_id=%d) \
[%s] (time=%.2f s): bandwidth=%.2fbit/s loss_rate=%.4f delay=%.4f ms", 
		    MIN_PIPE_ID_OUT + node_i, my_id, node_i, next_hop_id, 
		    IP_char_addresses + (node_i-FIRST_NODE_ID) * IP_ADDR_SIZE, 
		    crt_record_time, bandwidth, loss_rate, delay);
	    }
	  else
	    {
	      INFO ("-- Wireconf pipe=%d: #%d UCAST to #%d (next_hop_id=%d) \
[%s] (time=%.2f s): no valid record could be found => skipping configuration", 
		    MIN_PIPE_ID_OUT + node_i, my_id, node_i, next_hop_id, 
		    IP_char_addresses + (node_i-FIRST_NODE_ID) * IP_ADDR_SIZE, 
		    crt_record_time);
	      
	      bandwidth = 0.0; // equivalent to no limitation
	      delay = 0.0;
	      loss_rate = 0.0;
	    }
	  
	  if (configure_pipe (dummynet_socket_id, MIN_PIPE_ID_OUT + node_i,
			      bandwidth, delay, loss_rate, PIPE_QSIZE) == ERROR)
	    {
	      WARNING ("Error configuring UCAST pipe %d.", 
		       MIN_PIPE_ID_OUT + node_i);
	      exit(1);
	    }
	}			// end for loop unicast

      // config the broadcast links have source node id = i
      for (node_i = FIRST_NODE_ID; node_i < (node_count + FIRST_NODE_ID);
	   node_i++)
	{
	  if (node_i == my_id)
	    continue;

	  /* IGNORE FLAG FOR THE MOMENT

	  // if the parameters for the next hop did not change, there is no
	  // need to proceed any further
	  if (my_records_bcast_changed[node_i] == FALSE)
	  continue;
	  // otherwise, set the flag to FALSE and proceed 
	  else
	  my_records_bcast_changed[node_i] = FALSE;
	  */
	  
	  bandwidth = my_records_bcast[node_i].bandwidth;
	  delay = my_records_bcast[node_i].delay;
	  loss_rate = my_records_bcast[node_i].loss_rate;
#ifdef DEBUG
	  DEBUG ("Used contents of my_records_bcast for deltaQ parameters \
(index is node_i=%d).", node_i);
	  //io_binary_print_record (&(my_records_bcast[node_i]));
#endif

	  if (bandwidth != UNDEFINED_BANDWIDTH)
	    {
	      // print current configuration info
	      INFO ("-- Wireconf pipe=%d: #%d BCAST from #%d [%s] \
(time=%.2f s): bandwidth=%.2fbit/s loss_rate=%.4f delay=%.4f ms", 
		    MIN_PIPE_ID_IN_BCAST + node_i, my_id, node_i, 
		    IP_char_addresses + (node_i-FIRST_NODE_ID) * IP_ADDR_SIZE, 
		    crt_record_time, bandwidth, loss_rate, delay);
	    }
	  else
	    {
	      INFO ("-- Wireconf pipe=%d:  #%d BCAST from #%d [%s] \
(time=%.2f s): no valid record could be found => using no degradation", 
		    MIN_PIPE_ID_IN_BCAST + node_i, my_id, node_i, 
		    IP_char_addresses + (node_i-FIRST_NODE_ID) * IP_ADDR_SIZE, 
		    crt_record_time);

	      bandwidth = 0.0; // equivalent to no limitation
	      delay = 0.0;
	      loss_rate = 0.0;
	    }
	  if ( configure_pipe (dummynet_socket_id, 
			       MIN_PIPE_ID_IN_BCAST + node_i,
			       bandwidth, delay, loss_rate, PIPE_QSIZE) 
	       == ERROR)
	    {
	      WARNING ("Error configuring BCAST pipe %d.", 
		       MIN_PIPE_ID_IN_BCAST + node_i);
	      exit(1);
	    }

	}			// end for loop for broadcast
      // end config for broadcast traffic

      rule_count = MAX_RULE_COUNT;

      // get rules and print statistics for the current rule
      if (get_rules (dummynet_socket_id, &rule_data, &rule_data_alloc_size,
		     rules, &rule_count) == SUCCESS)
	{
	  int i, node_i;

	  long long unsigned int delta_pkt_counter, delta_byte_counter;

	  struct stats_class stats;

	  float adjusted_delta_pkt_counter;

	  my_total_channel_utilization = 0;
	  my_total_transmission_probability = 0;


	  wireconf.total_self_number_packets = 0;


	  node_i = 0;

	  // !!!!!!!!!!!!!!1 We assume there is one rule per node,
	  // and node ids and rule ids correspond !!!!!!!!!!!!!!!
	  for (i = 0; i < rule_count; i++)
	    {
	      // filter only the rules for UCAST
	      if (rules[i]->rulenum >= MIN_PIPE_ID_OUT
		  && rules[i]->rulenum < (MIN_PIPE_ID_OUT
					  + binary_header.interface_number))
		{
		  // check whether current node (node_i) is the same with
		  // the running node (my_id), case for which no records exist;
		  // increment current node index after setting values to 0
		  if (node_i == my_id)
		    {
		      wireconf.self_channel_utilizations[node_i] = 0;
		      wireconf.self_transmission_probabilities[node_i] = 0;
		      node_i++;
		    }

		  // because of check above, we know that node_i != my_id
		  DEBUG ("my_id=%d node_i=%d Rule #%d: \
\t%6"PRIu64" pkts \t%9"PRIu64" bytes", my_id, node_i, (rules[i])->rulenum, 
			 (rules[i])->pcnt, (rules[i])->bcnt);

		  delta_pkt_counter = rules[i]->pcnt
		    - last_pkt_counters[node_i];
		  delta_byte_counter = rules[i]->bcnt
		    - last_byte_counters[node_i];

		  last_pkt_counters[node_i] = rules[i]->pcnt;
		  last_byte_counters[node_i] = rules[i]->bcnt;

		  // update avg frame size statistics if packets exist
		  if (delta_pkt_counter > 0)
		    avg_frame_sizes[node_i] = (float) delta_byte_counter
		      / (float) delta_pkt_counter;

		  /*
		     wireconf.self_channel_utilizations[node_i] = 
		     compute_channel_utilization
		     (&(my_records_ucast[node_i]), delta_pkt_counter, 
		     delta_byte_counter);
		   */
		  wireconf.self_channel_utilizations[node_i] =
		    compute_channel_utilization
		    (&(adjusted_records_ucast[next_hop_ids[node_i]]),
		     delta_pkt_counter, delta_byte_counter,
		     0.5 /*!!!!!!!!!!!!!!!!!!!!! */ );


		  wireconf.self_transmission_probabilities[node_i] =
		    compute_transmission_probability
		    (&(adjusted_records_ucast[next_hop_ids[node_i]]),
		     delta_pkt_counter, delta_byte_counter,
		     0.5 /*!!!!!!!!!!!!!!!!!!!!! */ ,
		     &adjusted_delta_pkt_counter);

		  wireconf.total_self_number_packets
		    += adjusted_delta_pkt_counter;

		  DEBUG ("my_total_channel_utilization=%f", 
		     my_total_channel_utilization);

		  my_total_channel_utilization +=
		    wireconf.self_channel_utilizations[node_i];
		  my_total_transmission_probability +=
		    wireconf.self_transmission_probabilities[node_i];

		  // go to next node_i
		  node_i++;
		}
	    }			//end loop


	  DEBUG ("my_total_channel_utilization=%f \
my_total_transmission_probability=%f", my_total_channel_utilization, 
		my_total_transmission_probability);

	  if (my_total_channel_utilization > 1.0)
	    WARNING ("my_total_channel_utilization exceeds 1.0");
	  if (my_total_transmission_probability > 1.0)
	    WARNING ("my_total_transmission_probability exceeds 1.0");

	  stats.channel_utilization = my_total_channel_utilization;
	  stats.transmission_probability = my_total_transmission_probability;

	  // only send statistics if deltaQ adjustment is enabled;
	  // could also prevent above computation?!
#ifdef ADJUST_DELTAQ
	  // at this moment local experiments cannot use statistics
	  // maybe this can be fixed in "stats_send_thread" by using the 
	  // alias address for second interface?!
	  if (local_experiment == FALSE)
	    wireconf_deliver_stats (&wireconf, &stats);
#endif
	}
      else
	WARNING ("Cannot obtain rule information");

      // increase loop counter
      loop_count++;
    }


  ///////////////////////////////////////////////
  // End experiment

  // get the time at the end of the experiment
  gettimeofday (&tp_end, NULL);

  // print execution time
  INFO ("\nExperiment execution time=%.4f s",
	(tp_end.tv_sec + tp_end.tv_usec / 1.0e6) -
	(tp_begin.tv_sec + tp_begin.tv_usec / 1.0e6));

  // check unexpected state
  if (time_i >= binary_header.time_record_number)
    {
      INFO ("Finished reading configuration file");
      fflush (stdout);
    }
  else
    WARNING ("Execution finished before reading the entire configuration file");

  // if the loop was not entered not even once
  // there is an error
  if (loop_count == 0)
    WARNING ("No valid line was found for the node %d", my_id);
    

  ///////////////////////////////////////////////
  // Delete dummynet rules/pipes

  INFO ("Deleting rules and pipes from dummynet...");
  delete_all_rules_and_pipes(my_id, dummynet_socket_id, node_count, TRUE);

  // close dummynet socket
  INFO ("Closing dummynet socket...");
  close_socket (dummynet_socket_id);

  // finalize wireconf library
  if (wireconf_finalize (&wireconf) == ERROR)
    {
      WARNING ("Could not finalize wireconf library");
      exit (1);
    }

  // free allocated memory


  if (rule_data != NULL) // SUSPICIOUS; ALLOCATED BY GET_RULES
    free (rule_data);


  if (binary_records != NULL)
    free (binary_records);

  if (next_hop_ids != NULL)
    free (next_hop_ids);


  for (node_i = 0; node_i < binary_header.interface_number; node_i++)
    if (my_records_ucast[node_i] != NULL)
      free (my_records_ucast[node_i]);
  if (my_records_ucast != NULL)
    free (my_records_ucast);

  if (adjusted_records_ucast != NULL)
    free (adjusted_records_ucast);

  if (my_records_ucast_changed != NULL)
    free (my_records_ucast_changed);

  if (my_records_bcast != NULL)
    free (my_records_bcast);

  if (my_records_bcast_changed != NULL)
    free (my_records_bcast_changed);


  if (last_byte_counters != NULL)
    free (last_byte_counters);

  if (last_pkt_counters != NULL)
    free (last_pkt_counters);

  if (avg_frame_sizes != NULL)
    free (avg_frame_sizes);


  // close file descriptor
  fclose (qomet_fd);

  return 0;
}
