#!/bin/bash


# This generic script is used to validate test-specific calculations
# for a single test and scenario. It is not intended for direct use
# by users, who should insted use the specific scripts in each directory
# of the test suite.

# check arguments
if [ $# -lt 2 ]; then
    echo "Not enough parameters were supplied."
    echo "Usage: run_one_test.sh <qomet_scenario> <columns_to_filter> [<tab_count>]"
else
    # get scenario file name from command-line arguments
    QOMET_SCENARIO=$1

    # get the index of the columns to be filtered
    COLUMNS_TO_FILTER=$2

    # number of tabs before result
    if [ $# -eq 3 ]; then
	TABS=$3
    else
	TABS="\t\t\t"
    fi
fi

# program to run
DELTAQ=../../deltaQ/deltaQ

if [ ! -e ${DELTAQ} ]; then
    echo "ERROR: The program '${DELTAQ}' does not exist."
    exit 1
fi

# initialize other variables
QOMET_VALID_OUTPUT=${QOMET_SCENARIO}.out.valid
QOMET_NEW_OUTPUT=${QOMET_SCENARIO}.out
FILTERED_VALID_OUTPUT=/tmp/filtered_output.valid
FILTERED_NEW_OUTPUT=/tmp/filtered_output.new

# messages
SUCCESS_MESSAGE="Scenario '${QOMET_SCENARIO}'${TABS}SUCCESS"
FAILED_MESSAGE="Scenario '${QOMET_SCENARIO}'${TABS}FAILED"
TROUBLE_MESSAGE="ERROR: Scenario '${QOMET_SCENARIO}' caused trouble in 'diff'."

# run qomet on XML scenario
if [ "${DO_DEBUG}" == "true" ]; then
    ${DELTAQ} -t ${QOMET_SCENARIO}
else
    ${DELTAQ} -t ${QOMET_SCENARIO} &> /dev/null
fi    

# check that QOMET executed succefully before continuing
if [ $? -eq 0 ]; then
    # filter columns of interest from valid output file
    tail -n +3 ${QOMET_VALID_OUTPUT} 2> /dev/null | cut -d " " -f ${COLUMNS_TO_FILTER} > ${FILTERED_VALID_OUTPUT} 

    # filter columns of interest from new output file
    tail -n +3 ${QOMET_NEW_OUTPUT} 2> /dev/null | cut -d " " -f ${COLUMNS_TO_FILTER} > ${FILTERED_NEW_OUTPUT}

    # make diff between the filtered valid output and new output
    diff ${FILTERED_VALID_OUTPUT} ${FILTERED_NEW_OUTPUT} &> /dev/null

    DIFF_RESULT=$?
else
    echo "ERROR: '${DELTAQ}' returned an error."
    DIFF_RESULT=1
fi

# check result of diff
if [ ${DIFF_RESULT} -eq 0 ]; then
    # remove temporary file
    rm ${QOMET_NEW_OUTPUT}
    echo -e ${SUCCESS_MESSAGE}
elif [ ${DIFF_RESULT} -eq 1 ]; then
    echo -e ${FAILED_MESSAGE}
else
    echo -e ${TROUBLE_MESSAGE}
    DIFF_RESULT=1
fi

# remove temporary files
rm -f ${FILTERED_VALID_OUTPUT}
rm -f ${FILTERED_NEW_OUTPUT}

# return exit status
exit ${DIFF_RESULT}