
/*
 * Copyright (c) 2006-2013 The StarBED Project  All rights reserved.
 * 
 * See the file 'LICENSE' for licensing information.
 *
 */

/************************************************************************
 *
 * QOMET Emulator Implementation
 *
 * File name: test_ipfw_wrap.c
 * Function: Source file for testing ipfw_wrap.c.
 *
 * Authors: Razvan Beuran
 *
 *   $Revision: 128 $
 *   $LastChangedDate: 2009-02-06 10:21:50 +0900 (Fri, 06 Feb 2009) $
 *   $LastChangedBy: razvan $
 *
 ***********************************************************************/

#include <inttypes.h>

#include "wireconf.h"


////////////////////////////////////////////////////
// Test parameters
////////////////////////////////////////////////////

// parameters for test 1: simple IPs, egress
int RULE_ID1 = 4321;
int PIPE_ID1 = 8765;
char SRC_IP1[] = "40.30.20.10";
char DST_IP1[] = "80.70.60.50";
int DIRECTION1 = DIRECTION_OUT;
double BANDWIDTH1 = 1e6;
double DELAY1 = 11;
double PLR1 = 0;
double QSIZE1 = 50;

// parameters for test 2: from "any" to IP and port, ingress 
int RULE_ID2 = 4322;
int PIPE_ID2 = 8766;
char SRC_IP2[] = "any";
char DST_IP2[] = "80.70.60.60:92";
int DIRECTION2 = DIRECTION_IN;
double BANDWIDTH2 = 200e3;
double DELAY2 = 22;
double PLR2 = 0.2;
double QSIZE2 = 23;

// parameters for test 3: from IP and port to "any", both directions 
int RULE_ID3 = 4323;
int PIPE_ID3 = 8767;
char SRC_IP3[] = "40.30.20.30:93";
char DST_IP3[] = "any";
int DIRECTION3 = DIRECTION_BOTH;
double BANDWIDTH3 = 30e3;
double DELAY3 = 33;
double PLR3 = 0.03;
double QSIZE3 = 3;

// parameters for test 4: simple MACs, egress
int RULE_ID4 = 4324;
int PIPE_ID4 = 8768;
char SRC_MAC4[] = "60:50:40:30:20:10";
char DST_MAC4[] = "F0:E0:D0:C0:B0:A0";
int DIRECTION4 = DIRECTION_OUT;
double BANDWIDTH4 = 4e3;
double DELAY4 = 44;
double PLR4 = 0.04;
double QSIZE4 = 4;


////////////////////////////////////////////////////
// Other constants
////////////////////////////////////////////////////

unsigned SLEEP_TIME = 5;
int MAX_RULE_COUNT = 10;


////////////////////////////////////////////////////
// Helper functions
////////////////////////////////////////////////////

// helper function to add a rule and a pipe
int
do_add (int dummynet_socket_id, unsigned rule_id, unsigned pipe_id,
	char *src_addr, char *dst_addr, int direction, int is_ip_rule)
{
  if (is_ip_rule == TRUE)
    {
      INFO ("Adding IP rule '%d' and pipe '%d' from '%s' to '%s' \
(direction %s)...", rule_id, pipe_id, src_addr, dst_addr, (direction == DIRECTION_OUT) ? "OUT" : ((direction == DIRECTION_IN) ? "IN" : "BOTH"));
      if (add_rule_and_pipe (dummynet_socket_id, rule_id, pipe_id,
			     src_addr, dst_addr, direction) == ERROR)
	{
	  WARNING ("Could not add IP rule '%d' with pipe '%d' from '%s' to \
'%s' (direction %s).", rule_id, pipe_id, src_addr, dst_addr, (direction == DIRECTION_OUT) ? "OUT" : ((direction == DIRECTION_IN) ? "IN" : "BOTH"));
	  return ERROR;
	}
    }
  else
    {
      INFO ("Adding MAC rule '%d' and pipe '%d' from '%s' to '%s' \
(direction %s)...", rule_id, pipe_id, src_addr, dst_addr, (direction == DIRECTION_OUT) ? "OUT" : ((direction == DIRECTION_IN) ? "IN" : "BOTH"));
      if (add_rule_and_pipe_mac (dummynet_socket_id, rule_id, pipe_id,
				 src_addr, dst_addr, direction) == ERROR)
	{
	  WARNING ("Could not add MAC rule '%d' with pipe '%d' from '%s' to \
'%s' (direction %s).", rule_id, pipe_id, src_addr, dst_addr, (direction == DIRECTION_OUT) ? "OUT" : ((direction == DIRECTION_IN) ? "IN" : "BOTH"));
	  return ERROR;
	}
    }

  return SUCCESS;
}

// helper function to configure a pipe
int
do_configure (int dummynet_socket_id, unsigned pipe_id, float bandwidth,
	      float delay, float loss_rate, int queue_size)
{
  INFO ("Configuring pipe '%d': bandwidth=%.0f bps  delay=%.2f ms  \
loss_rate=%.2f  queue_size=%d", pipe_id, bandwidth, delay, loss_rate, queue_size);

  if (configure_pipe (dummynet_socket_id, pipe_id, bandwidth,
		      delay, loss_rate, queue_size) == ERROR)
    {
      WARNING ("Could not configure pipe '%d': bandwidth=%.0f bps  \
delay=%.2f ms  loss_rate=%.2f  queue_size=%d", pipe_id, bandwidth, delay, loss_rate, queue_size);
      return ERROR;
    }

  return SUCCESS;
}

// helper function to delete a rule and a pipe
int
do_delete (int dummynet_socket_id, unsigned rule_id, unsigned pipe_id)
{
  int status = SUCCESS;

  // delete a rule
  INFO ("Deleting rule '%d'...", rule_id);
  if (delete_rule (dummynet_socket_id, rule_id) == ERROR)
    {
      WARNING ("Could not delete rule '%d'.", rule_id);
      status = ERROR;
    }
  else
    INFO ("Delete rule operation succeeded.");

  // delete a pipe
  INFO ("Deleting pipe '%d'...", pipe_id);
  if (delete_pipe (dummynet_socket_id, pipe_id) == ERROR)
    {
      WARNING ("Could not delete pipe '%d'.", pipe_id);
      status = ERROR;
    }
  else
    INFO ("Delete pipe operation succeeded.");

  return status;
}


////////////////////////////////////////////////////
// Main function of the program
////////////////////////////////////////////////////

int
main ()
{
  // dummynet control socket id
  int dummynet_socket_id;

  // return status
  int status = SUCCESS;

  // memory to store rule data as a block
  struct ip_fw *rule_data = NULL;

  // size of allocate memory for rule data
  int rule_data_alloc_size = UNDEFINED_ALLOC_SIZE;

  // arrary of pointers to rule data
  struct ip_fw *rules[MAX_RULE_COUNT];

  // effective number of rules  
  int rule_count;

  // loop counter
  int i;


  ////////////////////////////////////////////////////
  // Test header

  // print settings
  INFO ("-----------------------------------------------------------");
  INFO ("Testing ipfw_wrap implementation in the wireconf library...");


  ////////////////////////////////////////////////////
  // Initial actions

  INFO ("-----------------------------------------------------------");
  INFO ("* INITIAL ACTIONS");

  // open dummynet socket
  INFO ("Opening Dummynet socket...");
  if ((dummynet_socket_id = get_socket ()) < 0)
    {
      WARNING ("Could not open Dummynet socket. NOTE: The test \
requires root privileges.");

      // fatal error, return directly
      return ERROR;
    }


  ////////////////////////////////////////////////////
  // Add rule and pipe operations

  INFO ("-----------------------------------------------------------");
  INFO ("* STEP 1: Adding rules and pipes ...");

  // add rule and pipe for test 1
  if (do_add (dummynet_socket_id, RULE_ID1, PIPE_ID1,
	      SRC_IP1, DST_IP1, DIRECTION1, TRUE) == ERROR)
    status = ERROR;

  // add rule and pipe for test 2
  if (do_add (dummynet_socket_id, RULE_ID2, PIPE_ID2,
	      SRC_IP2, DST_IP2, DIRECTION2, TRUE) == ERROR)
    status = ERROR;

  // add rule and pipe for test 3
  if (do_add (dummynet_socket_id, RULE_ID3, PIPE_ID3,
	      SRC_IP3, DST_IP3, DIRECTION3, TRUE) == ERROR)
    status = ERROR;

  // add rule and pipe for test 4
  if (do_add (dummynet_socket_id, RULE_ID4, PIPE_ID4,
	      SRC_MAC4, DST_MAC4, DIRECTION4, FALSE) == ERROR)
    status = ERROR;


  ////////////////////////////////////////////////////
  // Configure pipe operations

  INFO ("-----------------------------------------------------------");
  INFO ("* STEP 2: Configuring pipes ...");

  // configure pipe for test 1
  if (do_configure (dummynet_socket_id, PIPE_ID1,
		    BANDWIDTH1, DELAY1, PLR1, QSIZE1) == ERROR)
    status = ERROR;

  // configure pipe for test 2
  if (do_configure (dummynet_socket_id, PIPE_ID2,
		    BANDWIDTH2, DELAY2, PLR2, QSIZE2) == ERROR)
    status = ERROR;

  // configure pipe for test 3
  if (do_configure (dummynet_socket_id, PIPE_ID3,
		    BANDWIDTH3, DELAY3, PLR3, QSIZE3) == ERROR)
    status = ERROR;

  // configure pipe for test 4
  if (do_configure (dummynet_socket_id, PIPE_ID4,
		    BANDWIDTH4, DELAY4, PLR4, QSIZE4) == ERROR)
    status = ERROR;


  ////////////////////////////////////////////////////
  // Get rule and pipe information

  INFO ("-----------------------------------------------------------");
  INFO ("* STEP 3: Getting rules information ...");

  // set maximum number of rules
  rule_count = MAX_RULE_COUNT;

  // get all the rules and print statistics for them
  if (get_rules (dummynet_socket_id, &rule_data, &rule_data_alloc_size,
		 rules, &rule_count) == SUCCESS)
    for (i = 0; i < rule_count; i++)
      INFO ("Rule[%d]: rule_number=%u packet_count=%" PRIu64 " byte_count=%"
	    PRIu64 "", i, rules[i]->rulenum, rules[i]->pcnt, rules[i]->bcnt);
  else
    {
      WARNING ("Could not get rule and pipe information.");
      status = ERROR;
    }

  if (rule_data != NULL)
    free (rule_data);


  ////////////////////////////////////////////////////
  // Just sleeping for a while

  INFO ("Sleeping %u seconds...", SLEEP_TIME);
  sleep (SLEEP_TIME);


  ////////////////////////////////////////////////////
  // Delete rule and pipe operations

  INFO ("-----------------------------------------------------------");
  INFO ("* STEP 4: Deleting rules and pipes ...");

  // delete rule and pipe for test 1
  if (do_delete (dummynet_socket_id, RULE_ID1, PIPE_ID1) == ERROR)
    status = ERROR;

  // delete rule and pipe for test 2
  if (do_delete (dummynet_socket_id, RULE_ID2, PIPE_ID2) == ERROR)
    status = ERROR;

  // delete rule and pipe for test 3
  if (do_delete (dummynet_socket_id, RULE_ID3, PIPE_ID3) == ERROR)
    status = ERROR;

  // delete rule and pipe for test 4
  if (do_delete (dummynet_socket_id, RULE_ID4, PIPE_ID4) == ERROR)
    status = ERROR;


  ////////////////////////////////////////////////////
  // Final actions

  INFO ("-----------------------------------------------------------");
  INFO ("* FINAL ACTIONS");

  // close dummynet socket
  INFO ("Closing dummynet socket...");
  close_socket (dummynet_socket_id);

  // check whether any errors occurred
  if (status == ERROR)
    WARNING ("Errors encountered during execution. Make sure the ipfw kernel \
module is loaded (see README file for details).");

  return status;
}
